from collections.abc import Callable
from typing import Annotated, Any, TypeVar

import orjson
from pydantic import BeforeValidator, WrapSerializer
from pymatgen.apps.battery.conversion_battery import (
    ConversionElectrode,
    ConversionVoltagePair,
)
from pymatgen.apps.battery.insertion_battery import (
    InsertionElectrode,
    InsertionVoltagePair,
)
from typing_extensions import TypedDict

from emmet.core.types.enums import BatteryType
from emmet.core.types.pymatgen_types.balanced_reaction_adapter import (
    TypedBalancedReactionDict,
)
from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    TypedComputedEntryDict,
    TypedComputedStructureEntryDict,
    pop_cse_empty_keys,
)


class BaseVoltagePairDict(TypedDict):
    voltage: float
    frac_charge: float
    frac_discharge: float
    framework_formula: str
    mAh: float
    mass_charge: float
    mass_discharge: float
    vol_charge: float
    vol_discharge: float


class BaseInsertionVoltagePairDict(BaseVoltagePairDict):
    working_ion_entry: TypedComputedEntryDict


class BaseConversionVoltagePairDict(BaseVoltagePairDict):
    working_ion_entry: TypedComputedStructureEntryDict


class TypedInsertionVoltagePairDict(BaseInsertionVoltagePairDict):
    entry_charge: TypedComputedStructureEntryDict
    entry_discharge: TypedComputedStructureEntryDict


class TypedConversionVoltagePairDict(BaseConversionVoltagePairDict):
    entries_charge: list[TypedComputedStructureEntryDict]
    entries_discharge: list[TypedComputedStructureEntryDict]
    rxn: TypedBalancedReactionDict


TypedInsertionElectrodeDict = TypedDict(
    "TypedInsertionElectrodeDict",
    {
        "@module": str,
        "@class": str,
        "framework_formula": str,
        "stable_entries": list[TypedComputedStructureEntryDict],
        "unstable_entries": list[TypedComputedStructureEntryDict],
        "voltage_pairs": list[TypedInsertionVoltagePairDict],
        "working_ion_entry": TypedComputedEntryDict,
    },
)


TypedConversionElectrodeDict = TypedDict(
    "TypedConversionElectrodeDict",
    {
        "@module": str,
        "@class": str,
        "framework_formula": str,
        "initial_comp_formula": str,
        "voltage_pairs": list[TypedConversionVoltagePairDict],
        "working_ion_entry": TypedComputedStructureEntryDict,
    },
)


InsertionElectrodeTypeVar = TypeVar(
    "InsertionElectrodeTypeVar",
    InsertionElectrode,
    TypedInsertionElectrodeDict,
)

ConversionElectrodeTypeVar = TypeVar(
    "ConversionElectrodeTypeVar",
    ConversionElectrode,
    TypedConversionElectrodeDict,
)


def walk_voltage_pairs(voltage_pairs: list[dict[str, Any]], battery_type: BatteryType):
    pair: dict[str, Any]
    voltage_pair_cls: type[InsertionVoltagePair | ConversionVoltagePair]

    match battery_type:
        case BatteryType.insertion:
            voltage_pair_cls = InsertionVoltagePair
            for pair in voltage_pairs:
                for key in ["entry_charge", "entry_discharge", "working_ion_entry"]:
                    pair[key] = pop_cse_empty_keys(pair[key])

        case BatteryType.conversion:
            voltage_pair_cls = ConversionVoltagePair
            for pair in voltage_pairs:
                for key in ["entries_charge", "entries_discharge"]:
                    pair[key] = [pop_cse_empty_keys(cse) for cse in pair[key]]

                pair["working_ion_entry"] = pop_cse_empty_keys(
                    pair["working_ion_entry"]
                )

    return [voltage_pair_cls.from_dict(vp) for vp in voltage_pairs]


def electrode_object_energy_adjustments_serde(
    d: dict, battery_type: BatteryType, serde_fn: Callable, ea="energy_adjustments"
):
    pair: dict[str, Any]

    d["working_ion_entry"][ea] = serde_fn(d["working_ion_entry"][ea])

    match battery_type:
        case BatteryType.insertion:
            for pair in d["voltage_pairs"]:
                for key in ["working_ion_entry", "entry_charge", "entry_discharge"]:
                    pair[key][ea] = serde_fn(pair[key][ea])

            for key in ["stable_entries", "unstable_entries"]:
                for entry in d[key]:
                    entry[ea] = serde_fn(entry[ea])

        case BatteryType.conversion:
            for pair in d["voltage_pairs"]:
                pair["working_ion_entry"][ea] = serde_fn(pair["working_ion_entry"][ea])
                for key in ["entries_charge", "entries_discharge"]:
                    for entry in pair[key]:
                        entry[ea] = serde_fn(entry[ea])


def electrode_object_deserializer(
    eo: dict[str, Any] | InsertionElectrode | ConversionElectrode,
) -> InsertionElectrode | ConversionElectrode:
    if isinstance(eo, dict):
        target_class: type[InsertionElectrode | ConversionElectrode]

        match eo["@class"]:
            case "InsertionElectrode":
                target_class = InsertionElectrode
                battery_type = BatteryType.insertion
                eo["working_ion_entry"] = pop_cse_empty_keys(eo["working_ion_entry"])
                for key in ["stable_entries", "unstable_entries"]:
                    eo[key] = [pop_cse_empty_keys(cse) for cse in eo[key]]

            case "ConversionElectrode":
                target_class = ConversionElectrode
                battery_type = BatteryType.conversion
                eo["working_ion_entry"] = pop_cse_empty_keys(eo["working_ion_entry"])

        if isinstance(eo["working_ion_entry"].get("energy_adjustments"), str):
            electrode_object_energy_adjustments_serde(eo, battery_type, orjson.loads)

        eo["voltage_pairs"] = walk_voltage_pairs(eo["voltage_pairs"], battery_type)

        return target_class.from_dict(eo)

    return eo


def electrode_object_serializer(electrode_object, nxt, info) -> dict[str, Any]:
    default_serialized_object = nxt(electrode_object.as_dict(), info)

    format = info.context.get("format") if info.context else None
    if format == "arrow":
        battery_type = (
            BatteryType.insertion
            if default_serialized_object["@class"] == "InsertionElectrode"
            else BatteryType.conversion
        )

        electrode_object_energy_adjustments_serde(
            default_serialized_object, battery_type, orjson.dumps
        )

    return default_serialized_object


InsertionElectrodeType = Annotated[
    InsertionElectrodeTypeVar,
    BeforeValidator(electrode_object_deserializer),
    WrapSerializer(electrode_object_serializer, return_type=dict[str, Any]),
]

ConversionElectrodeType = Annotated[
    ConversionElectrodeTypeVar,
    BeforeValidator(electrode_object_deserializer),
    WrapSerializer(electrode_object_serializer, return_type=dict[str, Any]),
]



"""Core definition of a Materials Document"""

from __future__ import annotations

from pydantic import Field
from pydantic.json_schema import SkipJsonSchema

from emmet.core.material import BasePropertyMetadata
from emmet.core.types.pymatgen_types.structure_adapter import StructureType


class PropertyDoc(BasePropertyMetadata):
    """
    Base model definition for any singular materials property.

    This may contain any amount of structure metadata for the
    purpose of search. This is intended to be inherited and
    extended, not used directly
    """

    property_name: str

    structure: SkipJsonSchema[StructureType | None] = Field(
        None, description="The structure associated with this property.", exclude=True
    )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.core.interface import GrainBoundary
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.lattice_adapter import TypedLatticeDict
from emmet.core.types.pymatgen_types.sites_adapter import TypedSiteDict
from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

TypedGrainBoundaryDict = TypedDict(
    "TypedGrainBoundaryDict",
    {
        "@module": str,
        "@class": str,
        "lattice": TypedLatticeDict,
        "sites": list[TypedSiteDict],
        "init_cell": TypedStructureDict,
        "rotation_axis": list[int],
        "rotation_angle": float,
        "gb_plane": list[int],
        "join_plane": list[int],
        "vacuum_thickness": float,
        "ab_shift": list[float],
        "oriented_unit_cell": TypedStructureDict,
    },
)

GrainBoundaryTypeVar = TypeVar(
    "GrainBoundaryTypeVar", GrainBoundary, TypedGrainBoundaryDict
)


def pop_empty_gb_keys(gb: GrainBoundaryTypeVar) -> GrainBoundary:
    if isinstance(gb, dict):
        for key in ["init_cell", "oriented_unit_cell"]:
            gb[key] = pop_empty_structure_keys(gb[key])  # type: ignore[literal-required]

        for site in gb["sites"]:
            if "name" in site and not site["name"]:
                del site["name"]

            for key in ["properties", "species"]:
                for prop, val in list(site.get(key, {}).items()):
                    if val is None:
                        del site[key][prop]

        return GrainBoundary.from_dict(gb)  # type: ignore[arg-type]

    return gb


GrainBoundaryType = Annotated[
    GrainBoundaryTypeVar,
    BeforeValidator(pop_empty_gb_keys),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedGrainBoundaryDict
    ),
]



from hashlib import blake2b

import numpy as np
from pydantic import Field
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Molecule
from pymatgen.core.trajectory import Trajectory

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPculeID
from emmet.core.qchem.task import TaskDocument

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class ForcesDoc(PropertyDoc):
    property_name: str = "forces"

    forces: list[list[float]] = Field(..., description="Atomic forces (units: Ha/Bohr)")

    precise_forces: list[list[float] | None] = Field(
        default_factory=list,
        description="High-precision atomic forces (units: Ha/Bohr)",
    )

    pcm_forces: list[list[float]] | None = Field(
        None,
        description="Electrostatic atomic forces from polarizable continuum model (PCM) implicit solvation "
        "(units: Ha/Bohr).",
    )

    cds_forces: list[list[float]] | None = Field(
        None,
        description="Atomic force contributions from cavitation, dispersion, and structural rearrangement in the SMx "
        "family of implicit solvent models (units: Ha/Bohr)",
    )

    average_force_magnitude: float | None = Field(
        None, description="Average magnitude of atomic forces (units: Ha/Bohr)"
    )

    max_force_magnitude: float | None = Field(
        None,
        description="Maximum magnitude of atomic forces (units: Ha/Bohr)",
    )

    min_force_magnitude: float | None = Field(
        None,
        description="Minimum magnitude of atomic forces (units: Ha/Bohr)",
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct a force document from a task document

        :param task: document from which force properties can be extracted
        :param molecule_id: MPculeID
        :param deprecated: bool. Is this document deprecated?
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        if task.task_type.value != "Force":
            raise ValueError(
                "ForcesDoc can only be constructed from force calculations,"
                f"not {task.task_type.value}!"
            )

        mol = task.output.initial_molecule

        forces = task.output.gradients
        precise_forces = task.output.precise_gradients
        pcm_forces = task.output.pcm_gradients
        cds_forces = task.output.CDS_gradients

        calc = task.calcs_reversed[0]

        # Precise forces are either in output or don't exist
        # For PCM and CDS forces, can check "calcs_reversed"
        if pcm_forces is None:
            pcm_forces = calc.get("pcm_gradients")

        if cds_forces is None:
            cds_forces = calc.get("CDS_gradients")

        # Basic stats
        if precise_forces is not None:
            magnitudes = [np.linalg.norm(np.asarray(f)) for f in precise_forces]
        else:
            magnitudes = [np.linalg.norm(np.asarray(f)) for f in forces]

        average_force_magnitude = np.mean(magnitudes)
        max_force_magnitude = max(magnitudes)  # type: ignore[type-var]
        min_force_magnitude = min(magnitudes)  # type: ignore[type-var]

        id_string = f"forces-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            forces=forces,
            precise_forces=precise_forces,
            pcm_forces=pcm_forces,
            cds_forces=cds_forces,
            average_force_magnitude=average_force_magnitude,
            max_force_magnitude=max_force_magnitude,
            min_force_magnitude=min_force_magnitude,
            origins=[MolPropertyOrigin(name="forces", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )


class TrajectoryDoc(PropertyDoc):
    property_name: str = "optimization_trajectory"

    num_trajectories: int = Field(
        ...,
        description="Number of separate optimization trajectories extracted from this task",
    )

    species: list[Element | str] = Field(
        ...,
        description="Element or element name for each atom in the molecule for each optimization step",
    )

    geometries: list[list[list[list[float]]]] = Field(
        ...,
        description="XYZ positions of each atom in the molecule for each optimization step for each optimization "
        "trajectory (units: Angstrom)",
    )

    energies: list[list[float]] = Field(
        ...,
        description="Electronic energies for each optimization step for each optimization trajectory (units: Hartree)",
    )

    forces: list[list[list[list[float]]]] = Field(
        ...,
        description="Forces on each atom for each optimization step for each optimization trajectory (units: Ha/Bohr)",
    )

    pcm_forces: list[list[list[list[float]]] | None] = Field(
        default_factory=list,
        description="Electrostatic atomic forces from polarizable continuum model (PCM) implicit solvation "
        "for each optimization step for each optimization trajectory (units: Ha/Bohr).",
    )

    cds_forces: list[list[list[list[float]]] | None] = Field(
        default_factory=list,
        description="Atomic force contributions from cavitation, dispersion, and structural rearrangement in the SMx "
        "family of implicit solvent models, for each optimization step for each optimization trajectory "
        "(units: Ha/Bohr)",
    )

    mulliken_partial_charges: list[list[list[float]] | None] = Field(
        default_factory=list,
        description="Partial charges of each atom for each optimization step for each optimization trajectory, using "
        "the Mulliken method",
    )

    mulliken_partial_spins: list[list[list[float]] | None] = Field(
        default_factory=list,
        description="Partial spins of each atom for each optimization step for each optimization trajectory, using "
        "the Mulliken method",
    )

    resp_partial_charges: list[list[list[float]] | None] = Field(
        default_factory=list,
        description="Partial charges of each atom for each optimization step for each optimization trajectory, using "
        "the restrained electrostatic potential (RESP) method",
    )

    dipole_moments: list[list[list[float]] | None] = Field(
        default_factory=list,
        description="Molecular dipole moment for each optimization step for each optimization trajectory, "
        "(units: Debye)",
    )

    resp_dipole_moments: list[list[list[float]] | None] = Field(
        default_factory=list,
        description="Molecular dipole moment for each optimization step for each optimization trajectory, "
        "using the restrainted electrostatic potential (RESP) method (units: Debye)",
    )

    @property
    def molecules(self) -> list[list[Molecule]]:
        """
        Geometries along the optimization trajectory, represented as pymatgen Molecule objects.

        Args:
            self

        Returns:
            list[Molecule]: the optimization trajectory as a list of Molecules

        """

        return [
            [
                Molecule(
                    self.species,
                    geom,
                    charge=self.charge,
                    spin_multiplicity=self.spin_multiplicity,
                )
                for geom in trajectory
            ]
            for trajectory in self.geometries
        ]

    def as_trajectories(self) -> list[Trajectory]:
        """
        Represent this TrajectoryDoc as a list of pymatgen Trajectory objects.

        Args:
            self

        Returns:
            trajectories (list[Trajectory]): TrajectoryDoc represented as a collection of pymatgen Trajectory objects

        """

        trajectories = list()

        mol_trajectories = self.molecules

        for ii, mols in enumerate(mol_trajectories):
            num_steps = len(mols)

            # Frame (structure) properties
            frame_props = {"energies": self.energies[ii]}
            for prop in (
                "dipole_moments",
                "resp_dipole_moments",
            ):
                frame_props[prop] = []
                if (vals := getattr(self, prop, None)) is not None:
                    frame_props[prop] = vals[ii]

            # Site (atomic) properties
            site_props = {"forces": self.forces[ii]}
            for prop in (
                "pcm_forces",
                "cds_forces",
                "mulliken_partial_charges",
                "mulliken_partial_spins",
                "resp_partial_charges",
            ):
                site_props[prop] = []
                if (vals := getattr(self, prop, None)) is not None:
                    site_props[prop] = vals[ii]

            # Convert into a Trajectory object
            traj_frame_props = list()
            traj_mols = list()
            for jj in range(num_steps):
                step_dict = dict()

                for k, v in frame_props.items():
                    if v is not None:
                        step_dict[k] = v[jj]  # type: ignore

                step_mol = mols[jj]
                for k, v in site_props.items():  # type: ignore
                    if v is not None:
                        step_mol.add_site_property(property_name=k, values=v[jj])  # type: ignore[arg-type]

                traj_mols.append(step_mol)

                traj_frame_props.append(step_dict)

            traj = Trajectory.from_molecules(
                traj_mols, frame_properties=traj_frame_props, time_step=None
            )  # type: ignore[arg-type]
            trajectories.append(traj)

        return trajectories

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct a trajectory document from a task document

        :param task: document from which force properties can be extracted
        :param molecule_id: MPculeID
        :param deprecated: bool. Is this document deprecated?
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        if task.task_type.value not in [
            "Geometry Optimization",
            "Frequency Flattening Geometry Optimization",
            "Transition State Geometry Optimization",
            "Frequency Flattening Transition State Geometry Optimization",
        ]:
            raise ValueError(
                "TrajectoryDoc can only be constructed from geometry optimization calculations,"
                f"not {task.task_type.value}!"
            )

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        multiplicity = mol.spin_multiplicity

        species = None
        geometries = list()  # type: ignore
        energies = list()
        total_gradients = list()
        pcm_gradients = list()
        cds_gradients = list()
        mulliken_partial_charges = list()
        mulliken_partial_spins = list()
        resp_partial_charges = list()
        dipole_moments = list()
        resp_dipole_moments = list()

        for calculation in task.calcs_reversed:
            species = calculation.get("species", species)

            this_geometries = calculation.get("geometries")
            this_energies = calculation.get("energy_trajectory")
            this_total_gradients = calculation.get("gradients")
            this_pcm_gradients = calculation.get("pcm_gradients")
            this_cds_gradients = calculation.get("CDS_gradients")

            this_mulliken = calculation.get("Mulliken")
            this_resp = calculation.get("RESP")
            this_dipoles = calculation.get("dipoles")

            valid_trajectory = True
            if this_geometries is None or this_energies is None:
                # No valid geometry optimization found
                valid_trajectory = False
            elif len(this_energies) != len(this_total_gradients):
                # Energies and forces cannot be trivially mapped
                valid_trajectory = False
            elif len(this_geometries) != len(this_energies):
                # Initial geometry not included - common because of how parsing is done
                if len(this_geometries) == len(this_energies) - 1:
                    this_geometries = [
                        calculation["initial_geometry"]
                    ] + this_geometries
                # Other issue - no one-to-one mapping of molecule structure and energy
                else:
                    valid_trajectory = False

            if not valid_trajectory:
                continue

            if isinstance(calculation["initial_molecule"], Molecule):
                init_mol = calculation["initial_molecule"]  # type: ignore
            else:
                init_mol = Molecule.from_dict(calculation["initial_molecule"])  # type: ignore

            if species is None:
                species = init_mol.species  # type: ignore

            # Number of steps
            # All data in this (sub)-trajectory must have the same length
            num_steps = len(geometries)

            this_dipole_moments = None
            this_resp_dipole_moments = None

            # electric dipoles
            if this_dipoles is not None:
                if (
                    this_dipoles.get("dipole") is not None
                    and len(this_dipoles["dipole"]) > 0
                ):
                    if (
                        isinstance(this_dipoles["dipole"][0], list)
                        and len(this_dipoles["dipole"]) == num_steps
                    ):
                        this_dipole_moments = this_dipoles["dipole"]
                if (
                    this_dipoles.get("RESP_dipole") is not None
                    and len(this_dipoles["RESP_dipole"]) > 0
                ):
                    if (
                        isinstance(this_dipoles["RESP_dipole"][0], list)
                        and len(this_dipoles["RESP_dipole"]) == num_steps
                    ):
                        this_resp_dipole_moments = this_dipoles["RESP_dipole"]

            this_mulliken_partial_charges = None
            this_mulliken_partial_spins = None
            this_resp_partial_charges = None

            # Partial charges/spins
            if this_mulliken is not None:
                if len(this_mulliken) == num_steps:
                    if int(multiplicity) == 1:
                        this_mulliken_partial_charges = this_mulliken
                    else:
                        # For open-shell molecules, need to split mulliken charges and spins
                        charges = list()
                        spins = list()

                        for step in this_mulliken:
                            step_charges = list()
                            step_spins = list()
                            for atom in step:
                                step_charges.append(atom[0])
                                step_spins.append(atom[1])
                            charges.append(step_charges)
                            spins.append(step_spins)

                        this_mulliken_partial_charges = charges
                        this_mulliken_partial_spins = spins
                elif len(this_mulliken) == num_steps + 1:
                    last = np.asarray(this_mulliken[-1])
                    seclast = np.asarray(this_mulliken[-2])
                    if np.allclose(last, seclast):
                        if int(multiplicity) == 1:
                            this_mulliken_partial_charges = this_mulliken[:-1]

                        else:
                            charges = list()
                            spins = list()

                            for step in this_mulliken[::-1]:
                                step_charges = list()
                                step_spins = list()
                                for atom in step:
                                    step_charges.append(atom[0])
                                    step_spins.append(atom[1])
                                charges.append(step_charges)
                                spins.append(step_spins)

                            this_mulliken_partial_charges = charges
                            this_mulliken_partial_spins = spins

            if this_resp is not None:
                if len(this_resp) == num_steps:
                    this_resp_partial_charges = this_resp
                elif len(this_resp) == num_steps + 1:
                    last = np.asarray(this_resp[-1])
                    seclast = np.asarray(this_resp[-2])
                    if np.allclose(last, seclast):
                        this_resp_partial_charges = this_resp[:-1]

            geometries.append(this_geometries)
            energies.append(this_energies)
            total_gradients.append(this_total_gradients)
            pcm_gradients.append(this_pcm_gradients)
            cds_gradients.append(this_cds_gradients)
            mulliken_partial_charges.append(this_mulliken_partial_charges)
            mulliken_partial_spins.append(this_mulliken_partial_spins)
            resp_partial_charges.append(this_resp_partial_charges)
            dipole_moments.append(this_dipole_moments)
            resp_dipole_moments.append(this_resp_dipole_moments)

        num_trajectories = len(geometries)

        id_string = f"trajectory-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            num_trajectories=num_trajectories,
            species=species,
            geometries=geometries,
            energies=energies,
            forces=total_gradients,
            pcm_forces=pcm_gradients,
            cds_forces=cds_gradients,
            mulliken_partial_charges=mulliken_partial_charges,
            mulliken_partial_spins=mulliken_partial_spins,
            resp_partial_charges=resp_partial_charges,
            dipole_moments=dipole_moments,
            resp_dipole_moments=resp_dipole_moments,
            origins=[MolPropertyOrigin(name="trajectory", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )



from pydantic import BaseModel, Field
from typing_extensions import TypedDict


class EOSType(TypedDict):
    V0: float
    eos_energies: list[float]
    B: float
    C: float
    E0: float


class EOSDoc(BaseModel):
    """
    Fitted equations of state and energies and volumes used for fits.
    """

    energies: list[float] | None = Field(
        None,
        description="Common energies in eV/atom that the equations of state are plotted with.",
    )

    volumes: list[float] | None = Field(
        None,
        description="Common volumes in A³/atom that the equations of state are plotted with.",
    )

    eos: EOSType | None = Field(
        None,
        description="Data for each type of equation of state.",
    )

    material_id: str | None = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******.",
    )



"""Define extensions of the MPID class which are used throughout emmet."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field
from pymatgen.core import Element

from emmet.core.mpid import MPID, AlphaID
from emmet.core.types.enums import ThermoType
from emmet.core.types.typing import IdentifierType
from emmet.core.types.enums import XasType, XasEdge
from emmet.core.utils import type_override
from emmet.core.vasp.calc_types import RunType

if TYPE_CHECKING:
    from typing import Any
    from typing_extensions import Self


class SuffixedID(BaseModel):
    """Handle suffixed identifier formats (ex., thermo IDs)."""

    identifier: IdentifierType
    suffix: Enum
    separator: str = Field(default="_")

    @classmethod
    def from_str(cls, idx: str) -> Self:
        """Ensure the class can be instantiated from a string or dict."""
        sep = cls.model_fields["separator"].default
        parts = idx.split(sep, 1)
        return cls(
            identifier=parts[0],
            suffix=parts[1],  # type: ignore[arg-type]
        )

    def __str__(self) -> str:
        """Format as a string."""
        return self.separator.join([self.identifier.string, self.suffix.value])

    @classmethod
    def _deserialize(cls, v: Any) -> Self:
        """Utility to parse values into a suffixed ID from various formats."""
        if isinstance(v, cls):
            return v
        elif isinstance(v, str):
            return cls.from_str(v)
        elif isinstance(v, dict):
            return cls(**v)
        else:
            raise ValueError(f"{cls.__name__} cannot deserialize input type {type(v)}.")


class ThermoID(SuffixedID):
    """Thermodynamic data identifier."""

    suffix: ThermoType | RunType


class BatteryID(SuffixedID):
    """Identify battery / electrode data."""

    suffix: Element


@type_override({"suffix": str})
class XasSpectrumID(SuffixedID):
    suffix: tuple[XasType, Element, XasEdge]  # type: ignore[assignment]
    separator: str = Field(default="-")

    @classmethod
    def from_str(cls, idx: str) -> Self:
        """Ensure the class can be instantiated from a string or dict."""
        sep = cls.model_fields["separator"].default
        for i in range(2):
            if len(parts := idx.rsplit(sep, idx.count(sep) - i)) == 4:
                break
        return cls(
            identifier=parts[0],
            suffix=parts[1:],  # type: ignore[arg-type]
        )

    def __str__(self) -> str:
        """Format as a string."""
        return self.separator.join(
            [self.identifier.string, *(e.value for e in self.suffix)]
        )


def validate_identifier(
    idx: str | MPID | AlphaID | SuffixedID, serialize: bool = False
) -> str | MPID | AlphaID | SuffixedID:
    """Format an input string or identifier as a valid Materials Project format identifier.

    Parameters
    -----------
    idx : str or MPID or AlphaID or SuffixedID
        The input identifier, can either be an already instantiated
        identifier object, or a plain string
    serialize : bool = False
        If True, returns the string representation of the identifier object.
        If False, returns the object.
    """

    for id_cls in (AlphaID, *SuffixedID.__subclasses__()):
        try:
            parsed_idx = (
                AlphaID(idx).formatted
                if id_cls == AlphaID
                else id_cls._deserialize(idx)
            )
            break
        except Exception:
            continue
    else:
        raise ValueError(
            f"Invalid identifier {idx}, must be one of MPID, AlphaID, ThermoID, BatteryID, or XasSpectrumID."
        )

    if serialize:
        return (
            str(AlphaID(parsed_idx))
            if isinstance(parsed_idx, MPID | AlphaID)
            else str(parsed_idx)
        )
    return parsed_idx



from pydantic import BaseModel

from emmet.core.mpid import MPID, AlphaID
from emmet.core.types.typing import IdentifierType


def test_identifier_type():

    class TestClass(BaseModel):
        ID: IdentifierType

    tc = TestClass(ID="dogs")

    # ensure that extant MPIDs are deserialized to MPID,
    # and return AlphaID strings on model dump
    assert isinstance(tc.ID, MPID)
    assert tc.model_dump()["ID"].isalpha()
    assert not tc.model_dump()["ID"].isnumeric()

    tc = TestClass(ID=AlphaID._cut_point + 1)
    # ensure that new MPIDs are deserialized to MPID,
    # and return AlphaID strings on model dump
    assert isinstance(tc.ID, AlphaID)
    assert tc.model_dump()["ID"].isalpha()
    assert not tc.model_dump()["ID"].isnumeric()



"""Core definition of a Materials Document"""

# mypy: ignore-errors

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from pydantic import Field
from pymatgen.core.structure import Molecule

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.mpid import MPculeID
from emmet.core.qchem.calc_types import LevelOfTheory
from emmet.core.structure import MoleculeMetadata
from emmet.core.utils import utcnow

if TYPE_CHECKING:
    from typing_extensions import Self

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class PropertyDoc(MoleculeMetadata):
    """
    Base model definition for any singular molecule property. This may contain any amount
    of molecule metadata for the purpose of search
    This is intended to be inherited and extended not used directly
    """

    property_name: str

    property_id: str = Field(
        ..., description="The unique identifier of this property document."
    )

    molecule_id: MPculeID = Field(
        ...,
        description="The ID of the molecule, used as a reference across property documents."
        "This comes in the form of an MPculeID (or appropriately formatted string)",
    )

    deprecated: bool = Field(
        ...,
        description="Whether this property document is deprecated.",
    )

    deprecation_reasons: list[str] | None = Field(
        None,
        description="List of deprecation tags detailing why this document isn't valid",
    )

    level_of_theory: LevelOfTheory | None = Field(
        None, description="Level of theory used to generate this property document."
    )

    solvent: str | None = Field(
        None,
        description="String representation of the solvent "
        "environment used to generate this property document.",
    )

    lot_solvent: str | None = Field(
        None,
        description="String representation of the level of theory and solvent "
        "environment used to generate this property document.",
    )

    last_updated: datetime = Field(
        description="Timestamp for the most recent calculation update for this property",
        default_factory=utcnow,
    )

    origins: list[MolPropertyOrigin] = Field(
        [], description="Dictionary for tracking the provenance of properties"
    )

    warnings: list[str] = Field([], description="Any warnings related to this property")

    @classmethod
    def from_molecule(  # type: ignore[override]
        cls: Self,
        meta_molecule: Molecule,
        property_id: str,
        molecule_id: MPculeID,
        **kwargs,
    ) -> Self:
        """
        Builds a molecule document using the minimal amount of information
        """

        return super().from_molecule(
            meta_molecule=meta_molecule,
            property_id=property_id,
            molecule_id=molecule_id,
            **kwargs,
        )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.electronic_structure.bandstructure import Kpoint
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.lattice_adapter import MSONableTypedLatticeDict

TypedKpointDict = TypedDict(
    "TypedKpointDict",
    {
        "@module": str,
        "@class": str,
        "lattice": MSONableTypedLatticeDict,
        "fcoords": list[float],
        "ccoords": list[float],
        "label": str,
    },
)


KpointTypeVar = TypeVar("KpointTypeVar", Kpoint, TypedKpointDict)

KpointType = Annotated[
    KpointTypeVar,
    BeforeValidator(lambda x: Kpoint.from_dict(x) if isinstance(x, dict) else x),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedKpointDict),
]



from __future__ import annotations

from typing import TYPE_CHECKING

from monty.json import MontyDecoder
from pydantic import BaseModel, Field

from emmet.core.tasks import _VOLUMETRIC_FILES, TaskDoc
from emmet.core.types.pymatgen_types.defect_adapter import DefectType

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any


mdecoder = MontyDecoder().process_decoded


class DefectInfo(BaseModel):
    """Information related to a point defect."""

    defect_name: str = Field(
        title="The name of the defect",
    )

    bulk_formula: str = Field(
        title="Bulk Formula",
        description="Formula of the bulk structure.",
    )

    defect: DefectType = Field(
        title="Defect Object",
        description="Unit cell representation of the defect object.",
    )

    charge_state: int | None = Field(
        None,
        title="Charge State",
        description="Charge state of the defect.",
    )

    supercell_matrix: list[list[int, int, int]] | None = Field(  # type: ignore[type-arg]
        None,
        title="Supercell Matrix",
        description="Supercell matrix used to construct the defect supercell.",
    )


class DefectTaskDoc(DefectInfo, TaskDoc):
    """Defect Task Document.

    Contains all the task-level information for a defect supercell calculation.
    """

    @classmethod
    def from_directory(
        cls,
        dir_name: Path | str,
        volumetric_files: tuple[str, ...] = _VOLUMETRIC_FILES,
        additional_fields: dict[str, Any] | None = None,
        volume_change_warning_tol: float = 0.2,
        defect_info_key: str = "info",
        **vasp_calculation_kwargs,
    ) -> TaskDoc:
        """
        Create a task document from a directory containing VASP files.

        Parameters
        ----------
        dir_name
            The path to the folder containing the calculation outputs.
        store_additional_json
            Whether to store additional json files found in the calculation directory.
        volumetric_files
            Volumetric files to search for.
        additional_fields
            Dictionary of additional fields to add to output document.
        volume_change_warning_tol
            Maximum volume change allowed in VASP relaxations before the calculation is
            tagged with a warning.
        defect_info_key
            The key in the `additional_json` to extract the defect information from
        **vasp_calculation_kwargs
            Additional parsing options that will be passed to the
            :obj:`.Calculation.from_vasp_files` function.

        Returns
        -------
        TaskDoc
            A task document for the calculation.
        """
        tdoc = TaskDoc.from_directory(
            dir_name=dir_name,
            volumetric_files=volumetric_files,
            store_additional_json=True,
            additional_fields=additional_fields,
            volume_change_warning_tol=volume_change_warning_tol,
            **vasp_calculation_kwargs,
        )
        return cls.from_taskdoc(tdoc, defect_info_key=defect_info_key)

    @classmethod
    def from_taskdoc(
        cls,
        taskdoc: TaskDoc,
        defect_info_key: str = "info",
    ) -> DefectTaskDoc:
        """
        Create a DefectTaskDoc from a TaskDoc

        Args:
            taskdoc: TaskDoc to convert
            defect_info_key: The key in the `additional_json`
                to extract the defect information from

        Returns:
            DefectTaskDoc
        """
        additional_info = taskdoc.additional_json[defect_info_key]
        defect = additional_info["defect"]
        charge_state = additional_info["charge_state"]
        defect_name = additional_info["defect_name"]
        bulk_formula = additional_info["bulk_formula"]
        supercell_matrix = additional_info["sc_mat"]

        task_dict = taskdoc.model_dump()
        task_dict.update(
            {
                "defect_name": defect_name,
                "defect": defect,
                "charge_state": charge_state,
                "bulk_formula": bulk_formula,
                "supercell_matrix": supercell_matrix,
            }
        )

        return cls(**task_dict)



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.structure_matcher import StructureMatcher
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.species_adapter import MSONableTypedSpeciesDict

TypedComparatorDict = TypedDict(
    "TypedComparatorDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
    },
)  #  No concrete comparator class defines an as_dict() method, parent(Comparator) is abstract

TypedStructureMatcherDict = TypedDict(
    "TypedStructureMatcherDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
        "ltol": float,
        "stol": float,
        "angle_tol": float,
        "primitive_cell": bool,
        "scale": bool,
        "attempt_supercell": bool,
        "comparator": TypedComparatorDict,
        "supercell_size": str,
        "ignored_species": list[MSONableTypedSpeciesDict],
    },
)

StructureMatcherTypeVar = TypeVar(
    "StructureMatcherTypeVar", StructureMatcher, TypedStructureMatcherDict
)

StructureMatcherType = Annotated[
    StructureMatcherTypeVar,
    BeforeValidator(
        lambda x: StructureMatcher.from_dict(x) if isinstance(x, dict) else x
    ),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedStructureMatcherDict
    ),
]



import json

import pytest
from monty.io import zopen

from emmet.core import ARROW_COMPATIBLE
from emmet.core.tasks import TaskDoc
from emmet.core.vasp.calc_types import TaskType
from emmet.core.vasp.material import MaterialsDoc

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture
def test_tasks(test_dir):
    with zopen(test_dir / "test_si_tasks.json.gz") as f:
        tasks = json.load(f)

    for task in tasks:
        task.update({"is_valid": True})

    tasks = [TaskDoc(**t) for t in tasks]
    return tasks


def test_make_mat(test_tasks):
    material = MaterialsDoc.from_tasks(test_tasks)
    assert material.formula_pretty == "Si"
    assert len(material.task_ids) == 4
    assert len(material.entries.model_dump(exclude_none=True)) == 1

    bad_task_group = [
        task for task in test_tasks if task.task_type != TaskType.Structure_Optimization
    ]

    with pytest.raises(Exception):
        MaterialsDoc.from_tasks(bad_task_group, use_statics=False)


def test_make_deprecated_mat(test_tasks):
    bad_task_group = [
        task for task in test_tasks if task.task_type != TaskType.Structure_Optimization
    ]

    material = MaterialsDoc.construct_deprecated_material(bad_task_group)

    assert material.deprecated
    assert material.formula_pretty == "Si"
    assert len(material.task_ids) == 3
    assert material.entries is None


def test_schema():
    MaterialsDoc.schema()


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(test_tasks):
    doc = MaterialsDoc.from_tasks(test_tasks)
    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(MaterialsDoc)
    )
    test_arrow_doc = MaterialsDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert doc == test_arrow_doc



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from emmet.core.electronic_structure import BandstructureData, DosData
from emmet.core.material_property import PropertyDoc
from emmet.core.thermo import DecompositionProduct
from emmet.core.types.enums import ValueEnum, XasEdge, XasType
from emmet.core.types.pymatgen_types.element_adapter import ElementType
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.typing import IdentifierType

if TYPE_CHECKING:
    from typing_extensions import Self


class HasProps(ValueEnum):
    """
    Enum of possible hasprops values.
    """

    materials = "materials"
    thermo = "thermo"
    xas = "xas"
    grain_boundaries = "grain_boundaries"
    chemenv = "chemenv"
    electronic_structure = "electronic_structure"
    absorption = "absorption"
    bandstructure = "bandstructure"
    dos = "dos"
    magnetism = "magnetism"
    elasticity = "elasticity"
    dielectric = "dielectric"
    piezoelectric = "piezoelectric"
    surface_properties = "surface_properties"
    oxi_states = "oxi_states"
    provenance = "provenance"
    charge_density = "charge_density"
    eos = "eos"
    phonon = "phonon"
    insertion_electrodes = "insertion_electrodes"
    substrates = "substrates"


class SummaryStats(BaseModel):
    """
    Statistics about a specified SummaryDoc field.
    """

    field: str | None = Field(
        None,
        title="Field",
        description="Field name corresponding to a field in SummaryDoc.",
    )
    num_samples: int | None = Field(
        None,
        title="Sample",
        description="The number of documents sampled to generate statistics. "
        "If unspecified, statistics will be from entire database.",
    )
    min: float | None = Field(
        None,
        title="Minimum",
        description="The minimum value "
        "of the specified field used to "
        "generate statistics.",
    )
    max: float | None = Field(
        None,
        title="Maximum",
        description="The maximum value "
        "of the specified field used to "
        "generate statistics.",
    )
    median: float | None = Field(
        None, title="Median", description="The median of the field values."
    )
    mean: float | None = Field(
        None, title="Mean", description="The mean of the field values."
    )
    distribution: list[float] | None = Field(
        None,
        title="Distribution",
        description="List of floats specifying a kernel density "
        "estimator of the distribution, equally spaced "
        "between specified minimum and maximum values.",
    )


class XASSearchData(BaseModel):
    """
    Fields in XAS sub docs in summary
    """

    edge: XasEdge | None = Field(
        None,
        title="Absorption Edge",
        description="The interaction edge for XAS",
    )
    absorbing_element: ElementType | None = Field(
        None,
        description="Absorbing element.",
    )

    spectrum_type: XasType | None = Field(
        None,
        description="Type of XAS spectrum.",
    )


class GBSearchData(BaseModel):
    """
    Fields in grain boundary sub docs in summary
    """

    sigma: int | None = Field(
        None,
        description="Sigma value of the boundary.",
    )

    type: str | None = Field(
        None,
        description="Grain boundary type.",
    )

    gb_energy: float | None = Field(
        None,
        description="Grain boundary energy in J/m^2.",
    )

    rotation_angle: float | None = Field(
        None,
        description="Rotation angle in degrees.",
    )


class SummaryDoc(PropertyDoc):
    """
    Summary information about materials and their properties, useful for materials
    screening studies and searching.
    """

    property_name: str = "summary"

    # Materials
    task_ids: list[IdentifierType] = Field(
        [],
        title="Calculation IDs",
        description="List of Calculations IDs associated with this material.",
    )

    structure: StructureType = Field(
        ...,
        description="The lowest energy structure for this material.",
    )

    # Thermo

    uncorrected_energy_per_atom: float | None = Field(
        None,
        description="The total DFT energy of this material per atom in eV/atom.",
    )

    energy_per_atom: float | None = Field(
        None,
        description="The total corrected DFT energy of this material per atom in eV/atom.",
    )

    formation_energy_per_atom: float | None = Field(
        None,
        description="The formation energy per atom in eV/atom.",
    )

    energy_above_hull: float | None = Field(
        None,
        description="The energy above the hull in eV/Atom.",
    )

    is_stable: bool = Field(
        False,
        description="Flag for whether this material is on the hull and therefore stable.",
    )

    equilibrium_reaction_energy_per_atom: float | None = Field(
        None,
        description="The reaction energy of a stable entry from the neighboring equilibrium stable materials in eV."
        " Also known as the inverse distance to hull.",
    )

    decomposes_to: list[DecompositionProduct] | None = Field(
        None,
        description="List of decomposition data for this material. Only valid for metastable or unstable material.",
    )

    # XAS

    xas: list[XASSearchData] | None = Field(
        None,
        description="List of xas documents.",
    )

    # GB

    grain_boundaries: list[GBSearchData] | None = Field(
        None,
        description="List of grain boundary documents.",
    )

    # Electronic Structure

    band_gap: float | None = Field(
        None,
        description="Band gap energy in eV.",
    )

    cbm: float | None = Field(
        None,
        description="Conduction band minimum data.",
    )

    vbm: float | None = Field(
        None,
        description="Valence band maximum data.",
    )

    efermi: float | None = Field(
        None,
        description="Fermi energy in eV.",
    )

    is_gap_direct: bool | None = Field(
        None,
        description="Whether the band gap is direct.",
    )

    is_metal: bool | None = Field(
        None,
        description="Whether the material is a metal.",
    )

    es_source_calc_id: IdentifierType | None = Field(
        None,
        description="The source calculation ID for the electronic structure data.",
    )

    bandstructure: BandstructureData | None = Field(
        None,
        description="Band structure data for the material.",
    )

    dos: DosData | None = Field(
        None,
        description="Density of states data for the material.",
    )

    # DOS

    dos_energy_up: float | None = Field(
        None,
        description="Spin-up DOS band gap in eV.",
    )

    dos_energy_down: float | None = Field(
        None,
        description="Spin-down DOS band gap in eV.",
    )

    # Magnetism

    is_magnetic: bool | None = Field(
        None,
        description="Whether the material is magnetic.",
    )

    ordering: str | None = Field(
        None,
        description="Type of magnetic ordering.",
    )

    total_magnetization: float | None = Field(
        None,
        description="Total magnetization in μB.",
    )

    total_magnetization_normalized_vol: float | None = Field(
        None,
        description="Total magnetization normalized by volume in μB/Å³.",
    )

    total_magnetization_normalized_formula_units: float | None = Field(
        None,
        description="Total magnetization normalized by formula unit in μB/f.u. .",
    )

    num_magnetic_sites: int | None = Field(
        None,
        description="The number of magnetic sites.",
    )

    num_unique_magnetic_sites: int | None = Field(
        None,
        description="The number of unique magnetic sites.",
    )

    types_of_magnetic_species: list[ElementType] | None = Field(
        None,
        description="Magnetic specie elements.",
    )

    # Elasticity

    # k_voigt: float | None = Field(None, description="Voigt average of the bulk modulus.")

    # k_reuss: float | None = Field(None, description="Reuss average of the bulk modulus in GPa.")

    # k_vrh: float | None = Field(None, description="Voigt-Reuss-Hill average of the bulk modulus in GPa.")

    # g_voigt: float | None = Field(None, description="Voigt average of the shear modulus in GPa.")

    # g_reuss: float | None = Field(None, description="Reuss average of the shear modulus in GPa.")

    # g_vrh: float | None = Field(None, description="Voigt-Reuss-Hill average of the shear modulus in GPa.")

    bulk_modulus: dict[str, float] | None = Field(
        None,
        description="Voigt, Reuss, and Voigt-Reuss-Hill averages of the bulk modulus in GPa.",
    )

    shear_modulus: dict[str, float] | None = Field(
        None,
        description="Voigt, Reuss, and Voigt-Reuss-Hill averages of the shear modulus in GPa.",
    )

    universal_anisotropy: float | None = Field(None, description="Elastic anisotropy.")

    homogeneous_poisson: float | None = Field(None, description="Poisson's ratio.")

    # Dielectric and Piezo

    e_total: float | None = Field(
        None,
        description="Total dielectric constant.",
    )

    e_ionic: float | None = Field(
        None,
        description="Ionic contribution to dielectric constant.",
    )

    e_electronic: float | None = Field(
        None,
        description="Electronic contribution to dielectric constant.",
    )

    n: float | None = Field(
        None,
        description="Refractive index.",
    )

    e_ij_max: float | None = Field(
        None,
        description="Piezoelectric modulus.",
    )

    # Surface Properties

    weighted_surface_energy_EV_PER_ANG2: float | None = Field(
        None,
        description="Weighted surface energy in eV/Å².",
    )

    weighted_surface_energy: float | None = Field(
        None,
        description="Weighted surface energy in J/m².",
    )

    weighted_work_function: float | None = Field(
        None,
        description="Weighted work function in eV.",
    )

    surface_anisotropy: float | None = Field(
        None,
        description="Surface energy anisotropy.",
    )

    shape_factor: float | None = Field(
        None,
        description="Shape factor.",
    )

    has_reconstructed: bool | None = Field(
        None,
        description="Whether the material has any reconstructed surfaces.",
    )

    # Oxi States

    possible_species: list[str] | None = Field(
        None,
        description="Possible charged species in this material.",
    )

    # Has Props

    has_props: dict[str, bool] | None = Field(
        None,
        description="List of properties that are available for a given material.",
    )

    # Theoretical

    theoretical: bool = Field(
        True,
        description="Whether the material is theoretical.",
    )

    # External Database IDs

    database_IDs: dict[str, list[str]] = Field(
        {}, description="External database IDs corresponding to this material."
    )

    @classmethod
    def from_docs(
        cls, material_id: IdentifierType | None = None, **docs: dict[str, dict]
    ) -> Self:
        """Converts a bunch of summary docs into a SummaryDoc"""
        doc = _copy_from_doc(docs)

        # Reshape document for various sub-sections
        # Electronic Structure + Bandstructure + DOS
        if "bandstructure" in doc:
            if doc["bandstructure"] is not None and list(
                filter(lambda x: x is not None, doc["bandstructure"].values())
            ):
                doc["has_props"]["bandstructure"] = True
            else:
                del doc["bandstructure"]
        if "dos" in doc:
            if doc["dos"] is not None and list(
                filter(lambda x: x is not None, doc["dos"].values())
            ):
                doc["has_props"]["dos"] = True
            else:
                del doc["dos"]
        if "task_id" in doc:
            del doc["task_id"]

        return SummaryDoc(material_id=material_id, **doc)


# Key mapping
summary_fields: dict[str, list] = {
    HasProps(k).value: v
    for k, v in {
        "materials": [
            "nsites",
            "elements",
            "nelements",
            "composition",
            "composition_reduced",
            "formula_pretty",
            "formula_anonymous",
            "chemsys",
            "volume",
            "density",
            "density_atomic",
            "symmetry",
            "structure",
            "deprecated",
            "task_ids",
            "builder_meta",
        ],
        "thermo": [
            "uncorrected_energy_per_atom",
            "energy_per_atom",
            "formation_energy_per_atom",
            "energy_above_hull",
            "is_stable",
            "equilibrium_reaction_energy_per_atom",
            "decomposes_to",
        ],
        "xas": ["absorbing_element", "edge", "spectrum_type", "spectrum_id"],
        "grain_boundaries": [
            "gb_energy",
            "sigma",
            "type",
            "rotation_angle",
            "w_sep",
        ],
        "electronic_structure": [
            "band_gap",
            "efermi",
            "cbm",
            "vbm",
            "is_gap_direct",
            "is_metal",
            "bandstructure",
            "dos",
            "task_id",
        ],
        "magnetism": [
            "is_magnetic",
            "ordering",
            "total_magnetization",
            "total_magnetization_normalized_vol",
            "total_magnetization_normalized_formula_units",
            "num_magnetic_sites",
            "num_unique_magnetic_sites",
            "types_of_magnetic_species",
            "is_magnetic",
        ],
        "elasticity": [
            "bulk_modulus",
            "shear_modulus",
            "universal_anisotropy",
            "homogeneous_poisson",
        ],
        "dielectric": ["e_total", "e_ionic", "e_electronic", "n"],
        "piezoelectric": ["e_ij_max"],
        "surface_properties": [
            "weighted_surface_energy",
            "weighted_surface_energy_EV_PER_ANG2",
            "shape_factor",
            "surface_anisotropy",
            "weighted_work_function",
            "has_reconstructed",
        ],
        "oxi_states": ["possible_species"],
        "provenance": ["theoretical", "database_IDs"],
        "charge_density": [],
        "eos": [],
        "phonon": [],
        "absorption": [],
        "insertion_electrodes": [],
        "substrates": [],
        "chemenv": [],
    }.items()
}


def _copy_from_doc(doc):
    """Helper function to copy the list of keys over from amalgamated document"""
    has_props = {str(val.value): False for val in HasProps}
    d = {"has_props": has_props, "origins": []}
    # Complex function to grab the keys and put them in the root doc
    # if the item is a list, it makes one doc per item with those corresponding keys
    for doc_key in summary_fields:
        sub_doc = doc.get(doc_key, None)
        if isinstance(sub_doc, list) and len(sub_doc) > 0:
            d["has_props"][doc_key] = True
            d[doc_key] = []
            for sub_item in sub_doc:
                temp_doc = {
                    copy_key: sub_item[copy_key]
                    for copy_key in summary_fields[doc_key]
                    if copy_key in sub_item
                }
                d[doc_key].append(temp_doc)
        elif isinstance(sub_doc, dict):
            d["has_props"][doc_key] = True
            if sub_doc.get("origins", None):
                d["origins"].extend(sub_doc["origins"])
            d.update(
                {
                    copy_key: sub_doc[copy_key]
                    for copy_key in summary_fields[doc_key]
                    if copy_key in sub_doc
                }
            )
    return d



from typing_extensions import NotRequired, TypedDict

from emmet.core.types.pymatgen_types.properties import TypedSiteProperitesDict
from emmet.core.types.pymatgen_types.species_adapter import TypedSpeciesDict

MSONableTypedSiteDict = TypedDict(
    "MSONableTypedSiteDict",
    {
        "@class": str,
        "@module": str,
        "label": str,
        "name": str,
        "properties": TypedSiteProperitesDict,
        "species": list[TypedSpeciesDict],
        "abc": list[float, float, float],  # type: ignore[type-arg]
        "xyz": list[float, float, float],  # type: ignore[type-arg]
    },
)


class TypedSiteDict(TypedDict):
    label: NotRequired[str | None]
    name: NotRequired[str | None]
    properties: NotRequired[TypedSiteProperitesDict | None]
    species: NotRequired[list[TypedSpeciesDict] | None]
    abc: NotRequired[list[float, float, float] | None]  # type: ignore[type-arg]
    xyz: NotRequired[list[float, float, float] | None]  # type: ignore[type-arg]



from __future__ import annotations

from pydantic import BaseModel, Field

from emmet.core.mpid import MPID, MPculeID
from emmet.core.types.typing import DateTimeType


class MolPropertyOrigin(BaseModel):
    """Provenance for molecular properties, uses legacy MPID."""

    name: str = Field(..., description="The property name")

    task_id: MPID | MPculeID = Field(
        ..., description="The calculation ID this property comes from."
    )

    last_updated: DateTimeType = Field(  # type: ignore
        description="The timestamp when this calculation was last updated",
    )



from pydantic import BaseModel, Field
from emmet.core.types.enums import ValueEnum


class Value(BaseModel):
    min_value: float | None = Field(None, description="Minimal value.")
    max_value: float | None = Field(None, description="Maximal value.")
    values: list[float] = Field([], description="Enumerated values in the literature.")
    units: str = Field(..., description="Unit of this value.")


class Conditions(BaseModel):
    heating_temperature: list[Value] | None = Field(
        None, description="Heating temperatures."
    )
    heating_time: list[Value] | None = Field(None, description="Heating times.")
    heating_atmosphere: list[str] | None = Field(
        None, description="List of heating atmospheres."
    )
    mixing_device: str | None = Field(
        None, description="Mixing device, if this operation is MIXING."
    )
    mixing_media: str | None = Field(
        None, description="Mixing media, if this operation is MIXING."
    )


class OperationTypeEnum(str, ValueEnum):
    starting = "StartingSynthesis"
    mixing = "MixingOperation"
    shaping = "ShapingOperation"
    drying = "DryingOperation"
    heating = "HeatingOperation"
    quenching = "QuenchingOperation"


class Operation(BaseModel):
    type: OperationTypeEnum = Field(
        ..., description="Type of the operation as classified by the pipeline."
    )
    token: str = Field(
        ..., description="Token (word) of the operation as written in paper."
    )
    conditions: Conditions = Field(
        ..., description="The conditions linked to this operation."
    )



import numpy as np
from pytest import approx, fixture

from monty.serialization import loadfn
from pymatgen.core import Element, Structure, Molecule
from pymatgen.core.trajectory import Trajectory as PmgTraj

from emmet.core.tasks import TaskDoc
from emmet.core.testing_utils import DataArchive
from emmet.core.trajectory import Trajectory


@fixture(scope="module")
def si_task(test_dir):
    with DataArchive.extract(
        test_dir / "vasp" / "Si_old_double_relax.json.gz"
    ) as dir_name:
        return TaskDoc.from_directory(dir_name)


def test_reorder_sites():
    # Test that site ordering works correctly

    lattice = [[5.0, 0, 0], [4, 0, 0], [4.5, 0, 0]]
    coords = [
        [0, 0, 0],
        [0.0, 0.25, 0.0],
        [0.25, 0.0, 0.25],
        [0.0, 0.0, 0.25],
        [0.25, 0.0, 0.0],
        [0.25, 0.25, 0.25],
        [0.5, 0.5, 0.5],
        [0.75, 0.75, 0.75],
    ]

    elements = ["Si", "Si", "Ti", "N", "Ga", "As", "N", "Cu"]
    ref_z = [Element(ele).Z for ele in elements]

    # try a number of times to ensure ordering is always OK
    for _ in range(2 * len(elements)):
        new_z = [z for z in ref_z]
        np.random.shuffle(new_z)

        structure = Structure(lattice, [Element.from_Z(z) for z in new_z], coords)
        reordered_struct, reordered_idx = Trajectory.reorder_sites(structure, ref_z)
        assert all(site.specie.Z == ref_z[i] for i, site in enumerate(reordered_struct))
        assert all(
            np.abs(np.linalg.norm(reordered_struct[i].coords - structure[old_i].coords))
            < 1e-6
            for i, old_i in enumerate(reordered_idx)
        )

        # ensur that these also work for Molecules
        mol = Molecule.from_sites(structure)
        reordered_mol, reordered_idx = Trajectory.reorder_sites(structure, ref_z)
        assert all(site.specie.Z == ref_z[i] for i, site in enumerate(reordered_mol))
        assert all(
            np.abs(np.linalg.norm(reordered_mol[i].coords - mol[old_i].coords)) < 1e-6
            for i, old_i in enumerate(reordered_idx)
        )


def test_task_doc(si_task):
    traj = Trajectory.from_task_doc(si_task)[0]
    assert traj.num_ionic_steps == sum(
        len(cr.output.ionic_steps) for cr in si_task.calcs_reversed
    )
    istep = -1
    for cr in si_task.calcs_reversed:
        for ionic_step in cr.output.ionic_steps:
            assert traj.energy[istep] == approx(ionic_step.e_0_energy)
            istep -= 1


def test_parquet(si_task, tmp_dir):
    parqet_file = "test.parquet"

    traj = Trajectory.from_task_doc(si_task)[0]
    traj.to(parqet_file, compression="GZIP")

    new_traj = Trajectory.from_parquet(parqet_file)
    assert hash(new_traj) == hash(traj)


def test_pmg(si_task):
    traj = Trajectory.from_task_doc(si_task)[0]
    pmg_traj = traj.to(fmt="PMG")
    assert isinstance(pmg_traj, PmgTraj)
    assert len(pmg_traj) == traj.num_ionic_steps
    assert len(pmg_traj.frame_properties) == traj.num_ionic_steps

    # some float-rounding that happens here on round trip
    roundtrip = Trajectory.from_pmg(pmg_traj)

    for k in Trajectory.model_fields:
        if k == "ionic_step_properties":
            continue

        if (
            k
            in (
                "num_ionic_steps",
                "electronic_steps",
                "task_type",
                "run_type",
            )
            or getattr(traj, k, None) is None
        ):
            # Can't easily test for equality of electronic steps without bespoke code
            continue

        assert all(
            np.all(np.abs(np.array(new_val) - np.array(getattr(traj, k)[i])) < 1e-6)
            for i, new_val in enumerate(getattr(roundtrip, k))
        )


def test_mixed_calc_type(test_dir):
    # Test that Trajectory correctly creates new Trajectories for every
    # sequential calculation of different CalcType

    three_cr_task_dict = loadfn(test_dir / "mp-1120260_cr.json.gz")
    trajs = Trajectory.from_task_doc(TaskDoc(**three_cr_task_dict))
    assert len(trajs) == 2  # GGA static followed by two SCAN relaxes
    assert trajs[0].task_type.value == "Static"
    assert trajs[0].run_type.value == "GGA"

    assert trajs[1].task_type.value == "Structure Optimization"
    assert trajs[1].run_type.value == "SCAN"

    # now exchange order of calcs_reversed to get three trajectories
    three_cr_task_dict["calcs_reversed"] = [
        three_cr_task_dict["calcs_reversed"][idx] for idx in (0, 2, 1)
    ]
    trajs = Trajectory.from_task_doc(TaskDoc(**three_cr_task_dict))
    assert len(trajs) == 3  # SCAN relax -> GGA static -> SCAN relax
    assert trajs[1].task_type.value == "Static"
    assert trajs[1].run_type.value == "GGA"

    for idx in (
        0,
        2,
    ):
        assert trajs[idx].task_type.value == "Structure Optimization"
        assert trajs[idx].run_type.value == "SCAN"



import pytest
from tests.conftest import get_test_object

from emmet.core import ARROW_COMPATIBLE
from emmet.core.tasks import TaskDoc
from emmet.core.testing_utils import DataArchive, assert_schemas_equal
from emmet.core.vasp.task_valid import TaskState

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.mark.parametrize(
    "object_name",
    [
        pytest.param("SiOptimizeDouble", id="SiOptimizeDouble"),
        pytest.param("SiStatic", id="SiStatic"),
        pytest.param("SiNonSCFUniform", id="SiNonSCFUniform"),
    ],
)
def test_analysis_summary(test_dir, object_name):
    from monty.json import MontyDecoder, jsanitize

    from emmet.core.tasks import AnalysisDoc, _get_state
    from emmet.core.vasp.calculation import Calculation

    test_object = get_test_object(object_name)

    with DataArchive.extract(
        test_dir / "vasp" / f"{test_object.folder}.json.gz"
    ) as dir_name:
        calcs_reversed = []
        for task_name, files in test_object.task_files.items():
            doc, _ = Calculation.from_vasp_files(dir_name, task_name, **files)
            calcs_reversed.append(doc)
            # The 2 tasks of double-relaxation have been reversed in
            # "/atomate2/tests/vasp/schemas/conftest.py" for "SiOptimizeDouble"
            # task_files are in the order of {"relax2","relax1"}

    test_doc = AnalysisDoc.from_vasp_calc_docs(calcs_reversed)
    assert _get_state(calcs_reversed, test_doc) == TaskState.SUCCESS
    valid_doc = test_object.task_doc["analysis"]
    assert_schemas_equal(test_doc, valid_doc)

    # test document can be jsanitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    MontyDecoder().process_decoded(d)


@pytest.mark.parametrize(
    ("object_name", "task_name"),
    [
        pytest.param("SiOptimizeDouble", "relax1", id="SiOptimizeDouble"),
        pytest.param("SiStatic", "standard", id="SiStatic"),
        pytest.param("SiNonSCFUniform", "standard", id="SiNonSCFUniform"),
    ],
)
def test_input_summary(test_dir, object_name, task_name):
    from monty.json import MontyDecoder, jsanitize

    from emmet.core.vasp.calculation import Calculation

    test_object = get_test_object(object_name)

    files = test_object.task_files[task_name]
    with DataArchive.extract(
        test_dir / "vasp" / f"{test_object.folder}.json.gz"
    ) as dir_name:
        calc_doc, _ = Calculation.from_vasp_files(dir_name, task_name, **files)

    test_doc = calc_doc.input
    valid_doc = test_object.task_doc["input"]
    assert_schemas_equal(test_doc, valid_doc)

    # test document can be jsanitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    MontyDecoder().process_decoded(d)


@pytest.mark.parametrize(
    ("object_name", "task_name"),
    [
        pytest.param("SiOptimizeDouble", "relax2", id="SiOptimizeDouble"),
        pytest.param("SiStatic", "standard", id="SiStatic"),
        pytest.param("SiNonSCFUniform", "standard", id="SiNonSCFUniform"),
    ],
)
def test_output_summary(test_dir, object_name, task_name):
    from monty.json import MontyDecoder, jsanitize

    from emmet.core.tasks import OutputDoc
    from emmet.core.vasp.calculation import Calculation

    test_object = get_test_object(object_name)

    files = test_object.task_files[task_name]
    with DataArchive.extract(
        test_dir / "vasp" / f"{test_object.folder}.json.gz"
    ) as dir_name:
        calc_doc, _ = Calculation.from_vasp_files(dir_name, task_name, **files)

    test_doc = OutputDoc.from_vasp_calc_doc(calc_doc)
    valid_doc = test_object.task_doc["output"]
    assert_schemas_equal(test_doc, valid_doc)

    # test document can be jsanitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    MontyDecoder().process_decoded(d)


@pytest.mark.parametrize(
    "object_name",
    [
        pytest.param("SiOptimizeDouble", id="SiOptimizeDouble"),
        pytest.param("SiStatic", id="SiStatic"),
        pytest.param("SiNonSCFUniform", id="SiNonSCFUniform"),
    ],
)
def test_task_doc(test_dir, object_name, tmpdir):
    import os
    import shutil

    from monty.json import jsanitize
    from monty.serialization import dumpfn
    from pymatgen.alchemy.materials import TransformedStructure
    from pymatgen.entries.computed_entries import ComputedEntry
    from pymatgen.transformations.standard_transformations import (
        DeformStructureTransformation,
    )

    test_object = get_test_object(object_name)
    with DataArchive.extract(
        test_dir / "vasp" / f"{test_object.folder}.json.gz"
    ) as dir_name:
        test_doc = TaskDoc.from_directory(dir_name)

        assert_schemas_equal(test_doc, test_object.task_doc)

        # test document can be jsanitized
        jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

        # This is currently an issue as older versions of dumped custodian VaspJob objects are in the
        # test files. This needs to be updated to properly test decoding.
        # MontyDecoder().process_decoded(dct)

        # Test that additional_fields works
        test_doc = TaskDoc.from_directory(dir_name, additional_fields={"foo": "bar"})

        assert test_doc.model_dump()["foo"] == "bar"

        assert len(test_doc.calcs_reversed) == len(test_object.task_files)
        assert test_doc.state == TaskState.SUCCESS

        # ensure that number of electronic steps are correctly populated
        for cr in test_doc.calcs_reversed:
            assert len(cr.output.ionic_steps) == len(cr.output.num_electronic_steps)
            assert cr.output.num_electronic_steps == [
                len(ionic_step.electronic_steps) for ionic_step in cr.output.ionic_steps
            ]

        # ensure that run stats are not all identically zero (i.e., they are parsed correctly)
        for run_stats in test_doc.run_stats.values():
            assert any(abs(time) > 1e-6 for time in run_stats.model_dump().values())

        # Check that entry is populated when calcs_reversed is not None
        if test_doc.calcs_reversed:
            assert isinstance(
                test_doc.entry, ComputedEntry
            ), f"Unexpected entry {test_doc.entry} for {object_name}"

        # Test that transformations field works, using hydrostatic compression as example
        ts = TransformedStructure(
            test_doc.output.structure,
            transformations=[
                DeformStructureTransformation(
                    deformation=[
                        [0.9 if i == j else 0.0 for j in range(3)] for i in range(3)
                    ]
                )
            ],
        )
        ts_json = jsanitize(ts.as_dict())
        dumpfn(ts, f"{tmpdir}/transformations.json")
        for f in os.listdir(dir_name):
            if os.path.isfile(os.path.join(dir_name, f)):
                shutil.copy(os.path.join(dir_name, f), tmpdir)
        test_doc = TaskDoc.from_directory(tmpdir)
    # if other_parameters == {}, this is popped from the TaskDoc.transformations field
    # seems like @version is added by monty serialization
    # jsanitize needed because pymatgen.core.Structure.pbc is a tuple
    assert all(
        test_doc.transformations[k] == v
        for k, v in ts_json.items()
        if k
        not in (
            "other_parameters",
            "@version",
            "last_modified",
        )
    )
    assert isinstance(test_doc.transformations, dict)

    # now test case when transformations are serialized, relevant for atomate2
    test_doc = TaskDoc(
        **{
            "transformations": ts,
            **{
                k: v for k, v in test_doc.model_dump().items() if k != "transformations"
            },
        }
    )
    assert test_doc.transformations == ts


def test_lda_and_pseudo_format(test_dir, tmpdir):

    from emmet.core.tasks import TaskDoc

    with DataArchive.extract(
        test_dir / "vasp" / "lda_calc.json.gz",
    ) as dir_name:
        task = TaskDoc.from_directory(dir_name)

    assert task.run_type.name == "LDA"
    assert all(task.input.incar.get(k) is None for k in ("GGA", "METAGGA"))

    expected_pseudo = {
        "pot_type": "PAW",
        "functional": "LDA",
        "symbols": ["Si"],
    }

    assert all(
        getattr(task.input.pseudo_potentials, k) == v
        for k, v in expected_pseudo.items()
    )


def test_orig_inp_parsing(tmp_dir):
    """Test parsing of VASP input with variable suffix, like `.orig`."""

    from pathlib import Path

    from pymatgen.core import Structure
    from pymatgen.io.vasp import Incar, Kpoints

    from emmet.core.tasks import _parse_orig_inputs

    # demo simple cubic Copper structure
    structure = Structure(
        [[3 if i != j else 0.0 for j in range(3)] for i in range(3)],
        ["Cu"],
        [[0.0, 0.0, 0.0]],
    )

    incar = Incar.from_dict(
        {
            "ALGO": "Normal",
            "ENCUT": 600,
        }
    )
    kpoints = Kpoints()

    for suffix in (".orig", ".image", ".mangoes"):
        structure.to(f"POSCAR{suffix}")
        incar.write_file(f"INCAR{suffix}")
        kpoints.write_file(f"KPOINTS{suffix}")

        vi = _parse_orig_inputs(Path("."), suffix=suffix)
        assert all(k in vi for k in ("incar", "kpoints", "poscar"))
        assert len(vi) == 3


@pytest.mark.parametrize(
    "object_name",
    [
        pytest.param("SiOptimizeDouble", id="SiOptimizeDouble"),
        pytest.param("SiStatic", id="SiStatic"),
        pytest.param("SiNonSCFUniform", id="SiNonSCFUniform"),
    ],
)
@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(test_dir, object_name, tmpdir):
    test_object = get_test_object(object_name)

    with DataArchive.extract(
        test_dir / "vasp" / f"{test_object.folder}.json.gz"
    ) as dir_name:

        doc = TaskDoc.from_directory(dir_name)
        arrow_struct = pa.scalar(
            doc.model_dump(context={"format": "arrow"}), type=arrowize(TaskDoc)
        )

        # Avoiding comparisons of round tripped arrow doc vs. original doc
        # due to a few field types that get changed during json serialization
        # e.g., orig_inputs/input.kpoints.tet_weight is a float, but if the
        # val is 0 json dumps to int -> 0, 0 != 0.0 when comparing doc models
        # TBD on the value of getting this strict comparison correct

        assert TaskDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))



"""
Core module exposes the document interfaces
These will be ingested via Drones, built by Builders, and served via the API.
"""

from importlib.metadata import PackageNotFoundError, version
from importlib.util import find_spec

ARROW_COMPATIBLE = find_spec("pyarrow")

try:
    __version__ = version("emmet-core")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.lattice_adapter import MSONableTypedLatticeDict
from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)
from emmet.core.types.typing import TypedBandDict


class TypedBandGapDict(TypedDict):
    direct: bool
    transition: str
    energy: float


class TypedBranchDict(TypedDict):
    start_index: int
    end_index: int
    name: str


TypedBandDictureSymmLineDict = TypedDict(
    "TypedBandDictureSymmLineDict",
    {
        "@module": str,
        "@class": str,
        "lattice_rec": MSONableTypedLatticeDict,
        "efermi": float,
        "kpoints": list[list[float, float, float]],  # type: ignore[type-arg]
        "bands": dict[str, list[list[float]]],
        "is_metal": bool,
        "vbm": TypedBandDict,
        "cbm": TypedBandDict,
        "band_gap": TypedBandGapDict,
        "labels_dict": dict[str, list[float]],
        "is_spin_polarized": bool,
        "projections": dict[str, list[list[list[list[float]]]]],
        "structure": TypedStructureDict,
        "branches": list[TypedBranchDict],
    },
)

BandStructureSymmLineTypeVar = TypeVar(
    "BandStructureSymmLineTypeVar",
    BandStructureSymmLine,
    TypedBandDictureSymmLineDict,
)


def pop_empty_bs_keys(bs: BandStructureSymmLineTypeVar):
    if isinstance(bs, dict):
        bs["structure"] = pop_empty_structure_keys(bs["structure"])
        return BandStructureSymmLine.from_dict(bs)  # type: ignore[arg-type]

    return bs


BandStructureSymmLineType = Annotated[
    BandStructureSymmLineTypeVar,
    BeforeValidator(pop_empty_bs_keys),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedBandDictureSymmLineDict
    ),
]



"""Core definitions of a VASP calculation documents."""

from __future__ import annotations

import logging
import os
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import orjson
from monty.io import zopen
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from pymatgen.command_line.bader_caller import bader_analysis_from_path
from pymatgen.command_line.chargemol_caller import ChargemolAnalysis
from pymatgen.core.structure import Structure
from pymatgen.core.trajectory import Trajectory
from pymatgen.electronic_structure.bandstructure import BandStructure
from pymatgen.electronic_structure.core import OrbitalType
from pymatgen.electronic_structure.dos import CompleteDos, Dos
from pymatgen.io.vasp import BSVasprun, Kpoints, Locpot, Oszicar, Outcar, Poscar
from pymatgen.io.vasp import Potcar as VaspPotcar
from pymatgen.io.vasp import PotcarSingle, Vasprun, VolumetricData
from typing_extensions import NotRequired, TypedDict

from emmet.core.math import ListMatrix3D, Matrix3D, Vector3D
from emmet.core.types.enums import StoreTrajectoryOption, TaskState, VaspObject
from emmet.core.types.pymatgen_types.kpoints_adapter import KpointsType
from emmet.core.types.pymatgen_types.lattice_adapter import LatticeType
from emmet.core.types.pymatgen_types.outcar_adapter import OutcarType
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.utils import jsanitize, type_override
from emmet.core.vasp.calc_types import (
    CalcType,
    RunType,
    TaskType,
    calc_type,
    run_type,
    task_type,
)

if TYPE_CHECKING:
    from typing_extensions import Self


logger = logging.getLogger(__name__)


class Potcar(BaseModel):
    pot_type: str | None = Field(None, description="Pseudo-potential type, e.g. PAW")
    functional: str | None = Field(
        None, description="Functional type use in the calculation."
    )
    symbols: list[str] | None = Field(
        None, description="List of VASP potcar symbols used in the calculation."
    )


class CalculationBaseModel(BaseModel):
    """Wrapper around pydantic BaseModel with extra functionality."""

    def get(self, key: Any, default_value: Any | None = None) -> Any:
        return getattr(self, key, default_value)


class TypedStatisticsDict(TypedDict):
    MEAN: float
    ABSMEAN: float
    VAR: float
    MIN: float
    MAX: float


class TypedPotcarKeywordsDict(TypedDict):
    header: list[str]
    data: list[str]


class TypedPotcarStatsDict(TypedDict):
    header: TypedStatisticsDict
    data: TypedStatisticsDict


class TypedPotcarSummaryStatsDict(TypedDict):
    keywords: NotRequired[TypedPotcarKeywordsDict | None]
    stats: NotRequired[TypedPotcarStatsDict | None]


class PotcarSpec(BaseModel):
    """Document defining a VASP POTCAR specification."""

    titel: str | None = Field(None, description="TITEL field from POTCAR header")
    hash: str | None = Field(None, description="md5 hash of POTCAR file")
    summary_stats: TypedPotcarSummaryStatsDict | None = Field(
        None,
        description="summary statistics used to ID POTCARs without hashing",
    )

    @classmethod
    def from_potcar_single(cls, potcar_single: PotcarSingle) -> Self:
        """
        Get a PotcarSpec from a PotcarSingle.

        Parameters
        ----------
        potcar_single
            A potcar single object.

        Returns
        -------
        PotcarSpec
            A potcar spec.
        """
        return cls(
            titel=potcar_single.TITEL,
            hash=potcar_single.md5_header_hash,
            summary_stats=potcar_single._summary_stats,  # type: ignore[arg-type]
        )

    @classmethod
    def from_potcar(cls, potcar: VaspPotcar) -> list["PotcarSpec"]:
        """
        Get a list of PotcarSpecs from a Potcar.

        Parameters
        ----------
        potcar
            A potcar object.

        Returns
        -------
        list[PotcarSpec]
            A list of potcar specs.
        """
        return [cls.from_potcar_single(p) for p in potcar]

    @classmethod
    def from_file(cls, file_path: str | Path) -> list["PotcarSpec"]:
        """
        Get a list of PotcarSpecs from a Potcar.

        Parameters
        ----------
        file_path : str or Path of the POTCAR

        Returns
        -------
        list[PotcarSpec]
            A list of potcar specs.
        """
        if ".spec" in str(file_path):
            with zopen(file_path, "rt") as psf:
                return [cls(**ps) for ps in orjson.loads(psf.read())]
        return cls.from_potcar(VaspPotcar.from_file(str(file_path)))


@type_override({"incar": str, "parameters": str})
class CalculationInput(CalculationBaseModel):
    """Document defining VASP calculation inputs.

    Note that the following fields were formerly top-level fields on InputDoc,
    but are now properties of `CalculationInput`:
        pseudo_potentials (Potcar) : summary of the POTCARs used in the calculation
        xc_override (str) : the exchange-correlation functional used if not
            the one specified by POTCAR
        is_lasph (bool) : how the calculation set LASPH (aspherical corrections)
        magnetic_moments (list of floats) : on-site magnetic moments
    """

    incar: dict[str, Any] | None = Field(
        None, description="INCAR parameters for the calculation"
    )
    kpoints: KpointsType | None = Field(None, description="KPOINTS for the calculation")
    nkpoints: int | None = Field(None, description="Total number of k-points")
    potcar_spec: list[PotcarSpec] | None = Field(
        None, description="Title and hash of POTCAR files used in the calculation"
    )
    potcar: list[str] | None = Field(
        None, description="The symbols of the POTCARs used in the calculation."
    )
    potcar_type: list[str] | None = Field(
        None, description="List of POTCAR functional types."
    )
    parameters: dict | None = Field(None, description="Parameters from vasprun")
    lattice_rec: LatticeType | None = Field(
        None, description="Reciprocal lattice of the structure"
    )
    structure: StructureType | None = Field(
        None, description="Input structure for the calculation"
    )
    is_hubbard: bool = Field(
        default=False, description="Is this a Hubbard +U calculation"
    )
    hubbards: dict[str, float] | None = Field(
        None, description="The hubbard parameters used"
    )

    @field_validator("hubbards", mode="before")
    def hubbards_deserializer(cls, hubbards):
        if isinstance(hubbards, list):
            hubbards = {k: v for k, v in hubbards}

        return hubbards

    @field_serializer("incar", "parameters", mode="wrap")
    def incar_params_serializer(self, d, default_serializer, info):
        default_serialized_object = default_serializer(d, info)

        format = info.context.get("format") if info.context else None
        if format == "arrow":
            return orjson.dumps(default_serialized_object)

        return default_serialized_object

    @field_validator("incar", "parameters", mode="before")
    def incar_params_deserializer(cls, d):
        if isinstance(d, str):
            d = orjson.loads(d)

        if d and d == "parameters":
            d = {k.strip(): v for k, v in d.items()}

        return d

    @model_validator(mode="before")
    @classmethod
    def clean_inputs(cls, config: Any) -> Any:
        """Ensure whitespace in parameters and Kpoints are serialized.

        NOTE:
        ------
        A change in VASP introduced whitespace into some parameters,
        for example `<i type="string" name="GGA    ">PE</I>` was observed in
        VASP 6.4.3. This will lead to an incorrect return value from RunType.
        This validator will ensure that any already-parsed documents are fixed.
        """
        if (kpts := config.get("kpoints")) and isinstance(kpts, dict):
            config["kpoints"] = Kpoints.from_dict(kpts)

        # Issue: older tasks don't always contain `input.potcar`
        # and `input.potcar_spec`
        # Newer tasks contain `input.potcar`, `input.potcar_spec`,
        # and `orig_inputs.potcar_spec`

        # orig_inputs.potcar used to be what is now potcar_spec
        # try to coerce back here, but not guaranteed
        if (pcar := config.get("potcar")) and not config.get("potcar_spec"):
            if isinstance(pcar, dict):
                if pcar.get("@class") == "Potcar":
                    try:
                        # If POTCAR library is available
                        pcar = VaspPotcar.from_dict(pcar)
                    except ValueError:
                        # No POTCAR library available
                        config["potcar"] = pcar.get("symbols")
                        config["potcar_type"] = [pcar.get("functional")]
                else:
                    pcar = Potcar(**pcar)

            if isinstance(pcar, VaspPotcar):
                config["potcar_spec"] = PotcarSpec.from_potcar(pcar)
                config.pop("potcar")
            elif isinstance(pcar, Potcar):
                config["potcar_type"] = [pcar.pot_type] if pcar.pot_type else None
                config["potcar"] = pcar.symbols
            else:
                try:
                    pspec = [PotcarSpec(**x) for x in pcar]
                    config["potcar_spec"] = pspec
                    config.pop("potcar")
                except Exception:
                    if not isinstance(pcar, list) or not all(
                        isinstance(x, str) for x in pcar
                    ):
                        config.pop("potcar")

        if (pspec := config.get("potcar_spec")) and config.get("potcar") is None:
            config["potcar"] = []
            for spec in pspec:
                if isinstance(spec, PotcarSpec) and spec.titel:
                    titel = spec.titel
                elif isinstance(spec, dict) and spec.get("titel"):
                    titel = spec["titel"]
                else:
                    continue
                symbs = titel.split()
                if len(symbs) >= 2:
                    symb = symbs[1]
                else:
                    symb = symbs[0]
                config["potcar"].append(symb)

        return config

    @cached_property
    def poscar(self) -> Poscar | None:
        """Return pymatgen object representing the POSCAR file."""
        if self.structure:
            return Poscar(self.structure)
        return None

    @property
    def is_lasph(self) -> bool | None:
        """Report if the calculation was run with aspherical corrections.

        If `self.parameters` is populated, returns the value of `LASPH`
        from vasprun.xml, or its default value if not set (`False`).

        If `self.parameters` isn't populated (vasprun.xml wasn't parsed),
        returns `None`.
        """
        if self.parameters:
            return self.parameters.get("LASPH", False)
        return None

    @property
    def pseudo_potentials(self) -> Potcar | None:
        """Summarize the pseudo-potentials used."""
        if not self.potcar_type:
            return None

        if len(potcar_meta := self.potcar_type[0].split("_")) == 2:
            pot_type, func = potcar_meta
        elif len(potcar_meta) == 1:
            pot_type = potcar_meta[0]
            func = "LDA"

        return Potcar(pot_type=pot_type, functional=func, symbols=self.potcar)

    @property
    def xc_override(self) -> str | None:
        """Report the exchange-correlation functional used."""
        xc = None
        if self.incar:
            xc = self.incar.get("GGA") or self.incar.get("METAGGA")
        return xc.upper() if isinstance(xc, str) else xc

    @property
    def magnetic_moments(self) -> list[float] | None:
        """Report initial magnetic moments assigned to each atom."""
        return (self.parameters or {}).get("MAGMOM", None)

    @classmethod
    def from_vasprun(
        cls, vasprun: Vasprun, potcar_spec: list[PotcarSpec] | None = None
    ) -> Self:
        """
        Create a VASP input document from a Vasprun object.

        Parameters
        ----------
        vasprun
            A vasprun object.
        potcar_spec : list of dict
            If specified, the POTCAR spec to override that of vasprun.xml

        Returns
        -------
        CalculationInput
            The input document.
        """
        kpoints_dict = vasprun.kpoints.as_dict()
        kpoints_dict["actual_kpoints"] = [
            {"abc": list(k), "weight": w}
            for k, w in zip(vasprun.actual_kpoints, vasprun.actual_kpoints_weights)
        ]

        parameters = dict(vasprun.parameters).copy()
        incar = dict(vasprun.incar)
        if metagga := incar.get("METAGGA"):
            # Per issue #960, the METAGGA tag is populated in the
            # INCAR field of vasprun.xml, and not parameters
            parameters.update({"METAGGA": metagga})

        return cls(  # type: ignore[call-arg]
            structure=vasprun.initial_structure,
            incar=incar,
            kpoints=Kpoints.from_dict(kpoints_dict),
            nkpoints=len(kpoints_dict["actual_kpoints"]),
            potcar_spec=potcar_spec or [PotcarSpec(**ps) for ps in vasprun.potcar_spec],
            potcar_type=[s.split()[0] for s in vasprun.potcar_symbols],
            parameters=parameters,
            lattice_rec=vasprun.initial_structure.lattice.reciprocal_lattice,
            is_hubbard=vasprun.is_hubbard,
            hubbards=vasprun.hubbards,
        )


class RunStatistics(BaseModel):
    """Summary of the run statistics for a VASP calculation."""

    average_memory: float = Field(0, description="The average memory used in kb")
    max_memory: float = Field(0, description="The maximum memory used in kb")
    elapsed_time: float = Field(0, description="The real time elapsed in seconds")
    system_time: float = Field(0, description="The system CPU time in seconds")
    user_time: float = Field(
        0, description="The user CPU time spent by VASP in seconds"
    )
    total_time: float = Field(0, description="The total CPU time for this calculation")
    cores: int = Field(0, description="The number of cores used by VASP")

    @classmethod
    def from_outcar(cls, outcar: Outcar) -> Self:
        """
        Create a run statistics document from an Outcar object.

        Parameters
        ----------
        outcar
            An Outcar object.

        Returns
        -------
        RunStatistics
            The run statistics.
        """
        # rename these statistics
        mapping = {
            "Average memory used (kb)": "average_memory",
            "Maximum memory used (kb)": "max_memory",
            "Elapsed time (sec)": "elapsed_time",
            "System time (sec)": "system_time",
            "User time (sec)": "user_time",
            "Total CPU time used (sec)": "total_time",
            "cores": "cores",
        }
        run_stats: dict[str, int | float] = {}
        for k, v in mapping.items():
            stat = outcar.run_stats.get(k) or 0
            try:
                stat = float(stat)
            except ValueError:
                # sometimes the statistics are misformatted
                stat = 0

            run_stats[v] = stat

        return cls(**run_stats)  # type: ignore[arg-type]


class FrequencyDependentDielectric(BaseModel):
    """Frequency-dependent dielectric data."""

    real: list[list[float]] | None = Field(
        None,
        description="Real part of the frequency dependent dielectric constant, given at"
        " each energy as 6 components according to XX, YY, ZZ, XY, YZ, ZX",
    )
    imaginary: list[list[float]] | None = Field(
        None,
        description="Imaginary part of the frequency dependent dielectric constant, "
        "given at each energy as 6 components according to XX, YY, ZZ, XY, "
        "YZ, ZX",
    )
    energy: list[float] | None = Field(
        None,
        description="Energies at which the real and imaginary parts of the dielectric"
        "constant are given",
    )

    @classmethod
    def from_vasprun(cls, vasprun: Vasprun) -> Self:
        """
        Create a frequency-dependent dielectric calculation document from a vasprun.

        Parameters
        ----------
        vasprun : Vasprun
            A vasprun object.

        Returns
        -------
        FrequencyDependentDielectric
            A frequency-dependent dielectric document.
        """
        energy, real, imag = vasprun.dielectric
        return cls(real=real, imaginary=imag, energy=energy)


class ElectronPhononDisplacedStructures(BaseModel):
    """Document defining electron phonon displaced structures."""

    temperatures: list[float] | None = Field(
        None,
        description="The temperatures at which the electron phonon displacements "
        "were generated.",
    )
    structures: list[StructureType] | None = Field(
        None, description="The displaced structures corresponding to each temperature."
    )


class ElectronicStep(BaseModel):  # type: ignore
    """Document defining the information at each electronic step.

    Note, not all the information will be available at every step.
    """

    alphaZ: float | None = Field(None, description="The alpha Z term.")
    ewald: float | None = Field(None, description="The ewald energy.")
    hartreedc: float | None = Field(None, description="Negative Hartree energy.")
    XCdc: float | None = Field(None, description="Negative exchange energy.")
    pawpsdc: float | None = Field(
        None, description="Negative potential energy with exchange-correlation energy."
    )
    pawaedc: float | None = Field(None, description="The PAW double counting term.")
    eentropy: float | None = Field(None, description="The entropy (T * S).")
    bandstr: float | None = Field(
        None, description="The band energy (from eigenvalues)."
    )
    atom: float | None = Field(None, description="The atomic energy.")
    e_fr_energy: float | None = Field(None, description="The free energy.")
    e_wo_entrp: float | None = Field(None, description="The energy without entropy.")
    e_0_energy: float | None = Field(None, description="The internal energy.")

    model_config = ConfigDict(extra="allow")


class IonicStep(BaseModel):  # type: ignore
    """Document defining the information at each ionic step."""

    e_fr_energy: float | None = Field(None, description="The free energy.")
    e_wo_entrp: float | None = Field(None, description="The energy without entropy.")
    e_0_energy: float | None = Field(None, description="The internal energy.")
    forces: list[Vector3D] | None = Field(None, description="The forces on each atom.")
    stress: Matrix3D | None = Field(None, description="The stress on the lattice.")
    electronic_steps: list[ElectronicStep] | None = Field(
        None, description="The electronic convergence steps."
    )
    num_electronic_steps: int | None = Field(
        None, description="The number of electronic steps needed to reach convergence."
    )
    structure: StructureType | None = Field(
        None, description="The structure at this step."
    )

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def set_elec_step_count(self):
        if self.electronic_steps is not None:
            self.num_electronic_steps = len(self.electronic_steps)
        return self


class CoreCalculationOutput(BaseModel):
    """Document defining core VASP calculation outputs."""

    bandgap: float | None = Field(
        None, description="The band gap from the calculation in eV"
    )
    cbm: float | None = Field(
        None,
        description="The conduction band minimum in eV (if system is not metallic)",
    )
    density: float | None = Field(
        None, description="Density of final structure in units of g/cc."
    )
    direct_gap: float | None = Field(
        None, description="Direct band gap in eV (if system is not metallic)"
    )
    dos_properties: dict[str, dict[str, dict[str, float]]] | None = Field(
        None,
        description="Element- and orbital-projected band properties (in eV) for the "
        "DOS. All properties are with respect to the Fermi level.",
    )
    efermi: float | None = Field(
        None, description="The Fermi level from the calculation in eV"
    )
    energy: float | None = Field(
        None, description="The final total DFT energy for the calculation"
    )
    energy_per_atom: float | None = Field(
        None, description="The final DFT energy per atom for the calculation"
    )
    epsilon_ionic: ListMatrix3D | None = Field(
        None, description="The ionic part of the dielectric constant"
    )
    epsilon_static: ListMatrix3D | None = Field(
        None, description="The high-frequency dielectric constant"
    )
    epsilon_static_wolfe: ListMatrix3D | None = Field(
        None,
        description="The high-frequency dielectric constant w/o local field effects",
    )
    frequency_dependent_dielectric: FrequencyDependentDielectric | None = Field(
        None,
        description="Frequency-dependent dielectric information from an LOPTICS "
        "calculation",
    )
    is_gap_direct: bool | None = Field(
        None, description="Whether the band gap is direct"
    )
    is_metal: bool | None = Field(None, description="Whether the system is metallic")
    locpot: dict[str, list[float]] | None = Field(
        None, description="Average of the local potential along the crystal axes"
    )
    mag_density: float | None = Field(
        None,
        description="The magnetization density, defined as total_mag/volume "
        "(units of A^-3)",
    )
    optical_absorption_coeff: list[float] | None = Field(
        None, description="Optical absorption coefficient in cm^-1"
    )
    outcar: OutcarType | None = Field(
        None, description="Information extracted from the OUTCAR file"
    )
    structure: StructureType | None = Field(
        None, description="The final structure from the calculation"
    )
    transition: str | None = Field(
        None, description="Band gap transition given by CBM and VBM k-points"
    )
    vbm: float | None = Field(
        None, description="The valence band maximum in eV (if system is not metallic)"
    )

    @field_validator("dos_properties", mode="before")
    def dos_properties_deserializer(cls, dos_properties):
        if dos_properties and isinstance(dos_properties, list):
            dos_properties = {
                element: {
                    orbital: {key: value for key, value in property}
                    for orbital, property in properties
                }
                for element, properties in dos_properties
            }
        elif dos_properties and isinstance(next(iter(dos_properties.values())), list):
            dos_properties = {
                element: {
                    orbital: {key: value for key, value in property}
                    for orbital, property in properties
                }
                for element, properties in dos_properties.items()
            }

        return dos_properties


class CalculationOutput(CoreCalculationOutput):
    """Wrapper for CoreCalculationOutput for parsing and storing larger fields."""

    elph_displaced_structures: ElectronPhononDisplacedStructures | None = Field(
        None,
        description="Electron-phonon displaced structures, generated by setting "
        "PHON_LMC = True.",
    )
    force_constants: list[list[Matrix3D]] | None = Field(
        None, description="Force constants between every pair of atoms in the structure"
    )
    ionic_steps: list[IonicStep] | None = Field(
        None, description="Energy, forces, structure, etc. for each ionic step"
    )
    normalmode_eigenvals: list[float] | None = Field(
        None,
        description="Normal mode eigenvalues of phonon modes at Gamma. "
        "Note the unit changed between VASP 5 and 6.",
    )
    normalmode_eigenvecs: list[list[Vector3D]] | None = Field(
        None, description="Normal mode eigenvectors of phonon modes at Gamma"
    )
    normalmode_frequencies: list[float] | None = Field(
        None, description="Frequencies in THz of the normal modes at Gamma"
    )
    num_electronic_steps: list[int] | None = Field(
        None, description="The number of electronic steps in each ionic step."
    )
    run_stats: RunStatistics | None = Field(
        None, description="Summary of runtime statistics for this calculation"
    )

    @classmethod
    def from_vasp_outputs(
        cls,
        vasprun: Vasprun,
        outcar: Outcar | None,
        contcar: Poscar | None,
        locpot: Locpot | None = None,
        elph_poscars: list[Path] | None = None,
        store_trajectory: StoreTrajectoryOption | str = StoreTrajectoryOption.NO,
    ) -> Self:
        """
        Create a VASP output document from VASP outputs.

        Parameters
        ----------
        vasprun
            A Vasprun object.
        outcar
            An Outcar object.
        contcar
            A Poscar object.
        locpot
            A Locpot object.
        elph_poscars
            Path to displaced electron-phonon coupling POSCAR files generated using
            ``PHON_LMC = True``.
        store_trajectory
            Whether to store ionic steps as a pymatgen Trajectory object.
            Different value tune the amount of data from the ionic_steps
            stored in the Trajectory.
            If not NO, the `ionic_steps` field is left as None.
        Returns
        -------
            The VASP calculation output document.
        """
        try:
            bandstructure = vasprun.get_band_structure(efermi="smart")
            bandgap_info = bandstructure.get_band_gap()
            electronic_output = dict(
                efermi=bandstructure.efermi,
                vbm=bandstructure.get_vbm()["energy"],
                cbm=bandstructure.get_cbm()["energy"],
                bandgap=bandgap_info["energy"],
                is_gap_direct=bandgap_info["direct"],
                is_metal=bandstructure.is_metal(),
                direct_gap=bandstructure.get_direct_band_gap(),
                transition=bandgap_info["transition"],
            )
        except Exception:
            logger.warning("Error in parsing bandstructure")
            if vasprun.incar["IBRION"] == 1:
                logger.warning("VASP doesn't properly output efermi for IBRION == 1")
            electronic_output = {}

        freq_dependent_diel: FrequencyDependentDielectric | None = None
        try:
            freq_dependent_diel = FrequencyDependentDielectric.from_vasprun(vasprun)
        except KeyError:
            pass

        locpot_avg = None
        if locpot:
            locpot_avg = {
                str(i): locpot.get_average_along_axis(i).tolist() for i in range(3)
            }

        # parse force constants
        phonon_output = {}
        if hasattr(vasprun, "force_constants"):
            # convert eigenvalues to frequency
            eigs = -vasprun.normalmode_eigenvals
            frequencies = np.sqrt(np.abs(eigs)) * np.sign(eigs)

            # convert to THz in VASP 5 and lower; VASP 6 uses THz internally
            major_version = int(vasprun.vasp_version.split(".")[0])
            if major_version < 6:
                frequencies *= 15.633302

            phonon_output = dict(
                force_constants=vasprun.force_constants.tolist(),
                normalmode_frequencies=frequencies.tolist(),
                normalmode_eigenvals=vasprun.normalmode_eigenvals.tolist(),
                normalmode_eigenvecs=vasprun.normalmode_eigenvecs.tolist(),
            )

        if outcar and contcar:
            outcar_dict = outcar.as_dict()

            # use structure from CONTCAR as it is written to
            # greater precision than in the vasprun
            # but still need to copy the charge over
            structure = contcar.structure
            structure._charge = vasprun.final_structure._charge

            mag_density = (
                outcar.total_mag / structure.volume if outcar.total_mag else None
            )

            if len(outcar.magnetization) != 0:
                # patch calculated magnetic moments into final structure
                magmoms = [m["tot"] for m in outcar.magnetization]
                structure.add_site_property("magmom", magmoms)
        else:
            logger.warning(
                "No OUTCAR/CONTCAR available, some information will be missing from TaskDoc."
            )
            outcar_dict = {}
            structure = vasprun.final_structure
            mag_density = None

        # Parse DOS properties
        dosprop_dict = (
            _get_band_props(vasprun.complete_dos, structure)
            if hasattr(vasprun, "complete_dos") and vasprun.parameters["LORBIT"] >= 11
            else {}
        )

        elph_structures: ElectronPhononDisplacedStructures | None = None
        if elph_poscars is not None:
            elph_structures_dct: dict[str, list[Any]] = {
                "temperatures": [],
                "structures": [],
            }
            for elph_poscar in elph_poscars:
                temp = str(elph_poscar.name).replace("POSCAR.T=", "").replace(".gz", "")
                elph_structures_dct["temperatures"].append(temp)
                elph_structures_dct["structures"].append(
                    Structure.from_file(elph_poscar)
                )
            elph_structures = ElectronPhononDisplacedStructures(**elph_structures_dct)

        store_trajectory = StoreTrajectoryOption(store_trajectory)
        ionic_steps = (
            vasprun.ionic_steps
            if store_trajectory == StoreTrajectoryOption.NO
            else None
        )
        num_elec_steps = None
        if ionic_steps is not None:
            num_elec_steps = [
                len(ionic_step.get("electronic_steps", []) or [])
                for ionic_step in ionic_steps
            ]

        # The `epsilon_*` attrs of vasprun are now numpy arrays
        # and need to be explicitly checked if they are empty
        epsilons: dict[str, np.ndarray | list | None] = {}
        for attr in ("epsilon_static", "epsilon_static_wolfe", "epsilon_ionic"):
            if (v := getattr(vasprun, attr, None)) is not None and len(v) > 0:
                epsilons[attr] = v
            else:
                epsilons[attr] = None

        return cls(
            structure=structure,
            energy=vasprun.final_energy,
            energy_per_atom=vasprun.final_energy / len(structure),
            mag_density=mag_density,
            frequency_dependent_dielectric=freq_dependent_diel,
            elph_displaced_structures=elph_structures,
            dos_properties=dosprop_dict,
            ionic_steps=(
                [IonicStep(**step) for step in ionic_steps] if ionic_steps else None
            ),
            num_electronic_steps=num_elec_steps,
            locpot=locpot_avg,
            outcar=outcar_dict,
            run_stats=RunStatistics.from_outcar(outcar) if outcar else None,
            **epsilons,  # type: ignore[arg-type]
            **electronic_output,
            **phonon_output,
        )


@type_override({"bader": str, "ddec6": str})
class Calculation(CalculationBaseModel):
    """Full VASP calculation inputs and outputs."""

    dir_name: str | None = Field(
        None, description="The directory for this VASP calculation"
    )
    vasp_version: str | None = Field(
        None, description="VASP version used to perform the calculation"
    )
    has_vasp_completed: TaskState | None = Field(
        None, description="Whether VASP completed the calculation successfully"
    )
    input: CalculationInput | None = Field(
        None, description="VASP input settings for the calculation"
    )
    output: CalculationOutput | None = Field(
        None, description="The VASP calculation output"
    )
    completed_at: str | None = Field(
        None, description="Timestamp for when the calculation was completed"
    )
    task_name: str | None = Field(
        None, description="Name of task given by custodian (e.g., relax1, relax2)"
    )
    output_file_paths: dict[str, str] | None = Field(
        None,
        description="Paths (relative to dir_name) of the VASP output files "
        "associated with this calculation",
    )
    bader: dict | None = Field(None, description="Output from bader charge analysis")
    ddec6: dict | None = Field(None, description="Output from DDEC6 charge analysis")
    run_type: RunType | None = Field(
        None, description="Calculation run type (e.g., HF, HSE06, PBE)"
    )
    task_type: TaskType | None = Field(
        None, description="Calculation task type (e.g., Structure Optimization)."
    )
    calc_type: CalcType | None = Field(
        None, description="Return calculation type (run type + task_type)."
    )

    @field_validator("has_vasp_completed", mode="before")
    @classmethod
    def ensure_enum(cls, v: Any) -> TaskState:
        if isinstance(v, bool):
            v = TaskState.SUCCESS if v else TaskState.FAILED
        return v

    @field_serializer("bader", "ddec6", mode="wrap")
    def bader_ddec6__serializer(self, d, default_serializer, info):
        default_serialized_object = default_serializer(d, info)

        format = info.context.get("format") if info.context else None
        if format == "arrow":
            return orjson.dumps(jsanitize(default_serialized_object, allow_bson=True))

        return default_serialized_object

    @field_validator("bader", "ddec6", mode="before")
    def bader_ddec6_deserializer(cls, d):
        return orjson.loads(d) if isinstance(d, str) else d

    @field_validator("output_file_paths", mode="before")
    def output_fps_deserializer(cls, output_fps):
        if isinstance(output_fps, list):
            output_fps = {k: v for k, v in output_fps}

        return output_fps

    @classmethod
    def from_vasp_files(
        cls,
        dir_name: Path | str,
        task_name: str,
        vasprun_file: Path | str,
        outcar_file: Path | str,
        contcar_file: Path | str,
        volumetric_files: list[str] | None = None,
        elph_poscars: list[Path] | None = None,
        oszicar_file: Path | str | None = None,
        potcar_spec_file: Path | str | None = None,
        parse_dos: str | bool = False,
        parse_bandstructure: str | bool = False,
        average_locpot: bool = True,
        run_bader: bool = False,
        run_ddec6: bool | str = False,
        strip_bandstructure_projections: bool = False,
        strip_dos_projections: bool = False,
        store_volumetric_data: tuple[str] | None = None,
        store_trajectory: StoreTrajectoryOption | str = StoreTrajectoryOption.NO,
        vasprun_kwargs: dict | None = None,
    ) -> tuple["Calculation", dict[VaspObject, dict]]:
        """
        Create a VASP calculation document from a directory and file paths.

        Parameters
        ----------
        dir_name
            The directory containing the calculation outputs.
        task_name
            The task name.
        vasprun_file
            Path to the vasprun.xml file, relative to dir_name.
        outcar_file
            Path to the OUTCAR file, relative to dir_name.
        contcar_file
            Path to the CONTCAR file, relative to dir_name
        volumetric_files
            Path to volumetric files, relative to dir_name.
        elph_poscars
            Path to displaced electron-phonon coupling POSCAR files generated using
            ``PHON_LMC = True``, given relative to dir_name.
        oszicar_file
            Path to the OSZICAR file, relative to dir_name
        potcar_spec_file : Path | str | None = None
            Path to a POTCAR.spec file, relative to dir_name.
            Used in rehydration of a calculation from archived
            data, where the original POTCAR is not available.
        parse_dos
            Whether to parse the DOS. Can be:

            - "auto": Only parse DOS if there are no ionic steps (NSW = 0).
            - True: Always parse DOS.
            - False: Never parse DOS.

        parse_bandstructure
            How to parse the bandstructure. Can be:

            - "auto": Parse the bandstructure with projections for NSCF calculations
              and decide automatically if it's line or uniform mode.
            - "line": Parse the bandstructure as a line mode calculation with
              projections
            - True: Parse the bandstructure as a uniform calculation with
              projections .
            - False: Parse the band structure without projects and just store
              vbm, cbm, band_gap, is_metal and efermi rather than the full
              band structure object.

        average_locpot
            Whether to store the average of the LOCPOT along the crystal axes.
        run_bader : bool = False
            Whether to run bader on the charge density.
        run_ddec6 : bool or str = False
            Whether to run DDEC6 on the charge density. If a string, it's interpreted
            as the path to the atomic densities directory. Can also be set via the
            DDEC6_ATOMIC_DENSITIES_DIR environment variable. The files are available at
            https://sourceforge.net/projects/ddec/files.
        strip_dos_projections
            Whether to strip the element and site projections from the density of
            states. This can help reduce the size of DOS objects in systems with many
            atoms.
        strip_bandstructure_projections
            Whether to strip the element and site projections from the band structure.
            This can help reduce the size of DOS objects in systems with many atoms.
        store_volumetric_data
            Which volumetric files to store.
        store_trajectory
            Whether to store the ionic steps in a pymatgen Trajectory object and the
            amount of data to store from the ionic_steps. Can be:
            - FULL: Store the Trajectory. All the properties from the ionic_steps
              are stored in the frame_properties except for the Structure, to
              avoid redundancy.
            - PARTIAL: Store the Trajectory. All the properties from the ionic_steps
              are stored in the frame_properties except from Structure and
              ElectronicStep.
            - NO: Trajectory is not Stored.
            If not NO, :obj:'.CalculationOutput.ionic_steps' is set to None
            to reduce duplicating information.
        vasprun_kwargs
            Additional keyword arguments that will be passed to the Vasprun init.

        Returns
        -------
        Calculation
            A VASP calculation document.
        """
        dir_name = Path(dir_name)
        vasprun_file = dir_name / vasprun_file
        outcar_file = dir_name / outcar_file
        contcar_file = dir_name / contcar_file

        vasprun_kwargs = vasprun_kwargs if vasprun_kwargs else {}
        volumetric_files = [] if volumetric_files is None else volumetric_files
        vasprun = Vasprun(vasprun_file, **vasprun_kwargs)
        outcar = Outcar(outcar_file)
        if (
            os.path.getsize(contcar_file) == 0
            and vasprun.parameters.get("NELM", 60) == 1
        ):
            contcar = Poscar(vasprun.final_structure)
        else:
            contcar = Poscar.from_file(contcar_file)
        completed_at = str(datetime.fromtimestamp(vasprun_file.stat().st_mtime))

        output_file_paths = _get_output_file_paths(volumetric_files)
        vasp_objects: dict[VaspObject, Any] = _get_volumetric_data(
            dir_name, output_file_paths, store_volumetric_data
        )

        if (dos := _parse_dos(parse_dos, vasprun)) is not None:
            if strip_dos_projections:
                dos = Dos(dos.efermi, dos.energies, dos.densities)  # type: ignore[arg-type]
            vasp_objects[VaspObject.DOS] = dos  # type: ignore

        bandstructure = _parse_bandstructure(parse_bandstructure, vasprun)
        if bandstructure is not None:
            if strip_bandstructure_projections:
                bandstructure.projections = {}
            vasp_objects[VaspObject.BANDSTRUCTURE] = bandstructure  # type: ignore

        bader = None
        if run_bader and VaspObject.CHGCAR in output_file_paths:
            suffix = "" if task_name == "standard" else f".{task_name}"
            bader = bader_analysis_from_path(str(dir_name), suffix=suffix)

        ddec6 = None
        if run_ddec6 and VaspObject.CHGCAR in output_file_paths:
            densities_path = run_ddec6 if isinstance(run_ddec6, (str, Path)) else None
            ddec6 = ChargemolAnalysis(
                path=dir_name, atomic_densities_path=densities_path
            ).summary

        locpot = None
        if average_locpot:
            if VaspObject.LOCPOT in vasp_objects:
                locpot = vasp_objects[VaspObject.LOCPOT]  # type: ignore
            elif VaspObject.LOCPOT in output_file_paths:
                locpot_file = output_file_paths[VaspObject.LOCPOT]  # type: ignore
                locpot = Locpot.from_file(dir_name / locpot_file)

        potcar_spec: list[PotcarSpec] | None = None
        if potcar_spec_file:
            potcar_spec = PotcarSpec.from_file(potcar_spec_file)
        input_doc = CalculationInput.from_vasprun(vasprun, potcar_spec=potcar_spec)

        store_trajectory = StoreTrajectoryOption(store_trajectory)
        output_doc = CalculationOutput.from_vasp_outputs(
            vasprun,
            outcar,
            contcar,
            locpot=locpot,
            elph_poscars=elph_poscars,
            store_trajectory=store_trajectory,
        )
        if store_trajectory != StoreTrajectoryOption.NO:
            exclude_from_trajectory = set(["structure"])
            if store_trajectory == StoreTrajectoryOption.PARTIAL:
                exclude_from_trajectory.add("electronic_steps")
            frame_properties = [
                IonicStep(**x).model_dump(exclude=exclude_from_trajectory)
                for x in vasprun.ionic_steps
            ]
            if oszicar_file:
                try:
                    oszicar = Oszicar(dir_name / oszicar_file)
                    if "T" in oszicar.ionic_steps[0]:
                        for frame_property, oszicar_is in zip(
                            frame_properties, oszicar.ionic_steps
                        ):
                            frame_property["temperature"] = oszicar_is.get("T")
                except ValueError:
                    # there can be errors in parsing the floats from OSZICAR
                    pass

            traj = Trajectory.from_structures(
                [d["structure"] for d in vasprun.ionic_steps],
                frame_properties=frame_properties,
                constant_lattice=False,
            )
            vasp_objects[VaspObject.TRAJECTORY] = traj  # type: ignore

        # MD run
        if vasprun.parameters.get("IBRION", -1) == 0:
            if vasprun.parameters.get("NSW", 0) == vasprun.md_n_steps:
                has_vasp_completed = TaskState.SUCCESS
            else:
                has_vasp_completed = TaskState.FAILED
        # others
        else:
            has_vasp_completed = (
                TaskState.SUCCESS if vasprun.converged else TaskState.FAILED
            )

        return (
            cls(
                dir_name=str(dir_name),
                task_name=task_name,
                vasp_version=vasprun.vasp_version,
                has_vasp_completed=has_vasp_completed,
                completed_at=completed_at,
                input=input_doc,
                output=output_doc,
                output_file_paths={
                    k.name.lower(): v for k, v in output_file_paths.items()
                },
                bader=bader,
                ddec6=ddec6,
                run_type=run_type(input_doc.parameters),
                task_type=task_type(input_doc.model_dump()),
                calc_type=calc_type(input_doc.model_dump(), input_doc.parameters),
            ),
            vasp_objects,
        )

    @classmethod
    def from_vasprun(
        cls,
        path: Path | str,
        task_name: str = "Unknown vapsrun.xml",
        vasprun_kwargs: dict | None = None,
    ) -> Self:
        """
        Create a VASP calculation document from a directory and file paths.

        Parameters
        ----------
        path
            Path to the vasprun.xml file.
        task_name
            The task name.
        vasprun_kwargs
            Additional keyword arguments that will be passed to the Vasprun init.

        Returns
        -------
        Calculation
            A VASP calculation document.
        """
        path = Path(path)
        vasprun_kwargs = vasprun_kwargs if vasprun_kwargs else {}
        vasprun = Vasprun(path, **vasprun_kwargs)

        completed_at = str(datetime.fromtimestamp(path.stat().st_mtime))

        input_doc = CalculationInput.from_vasprun(vasprun)

        output_doc = CalculationOutput.from_vasp_outputs(
            vasprun,
            outcar=None,
            contcar=None,
        )

        # MD run
        if vasprun.parameters.get("IBRION", -1) == 0:
            if vasprun.parameters.get("NSW", 0) == vasprun.nionic_steps:
                has_vasp_completed = TaskState.SUCCESS
            else:
                has_vasp_completed = TaskState.FAILED
        # others
        else:
            has_vasp_completed = (
                TaskState.SUCCESS if vasprun.converged else TaskState.FAILED
            )

        return cls(  # type: ignore[call-arg]
            dir_name=str(path.resolve().parent),
            task_name=task_name,
            vasp_version=vasprun.vasp_version,
            has_vasp_completed=has_vasp_completed,
            completed_at=completed_at,
            input=input_doc,
            output=output_doc,
            output_file_paths={},
            run_type=run_type(input_doc.parameters),
            task_type=task_type(input_doc.model_dump()),
            calc_type=calc_type(input_doc.model_dump(), input_doc.parameters),
        )


def _get_output_file_paths(volumetric_files: list[str]) -> dict[VaspObject, str]:
    """
    Get the output file paths for VASP output files from the list of volumetric files.

    Parameters
    ----------
    volumetric_files
        A list of volumetric files associated with the calculation.

    Returns
    -------
    dict[VaspObject, str]
        A mapping between the VASP object type and the file path.
    """
    output_file_paths = {}
    for vasp_object in VaspObject:  # type: ignore
        for volumetric_file in volumetric_files:
            if vasp_object.name in str(volumetric_file):
                output_file_paths[vasp_object] = str(volumetric_file)
    return output_file_paths


def _get_volumetric_data(
    dir_name: Path,
    output_file_paths: dict[VaspObject, str],
    store_volumetric_data: tuple[str] | None,
) -> dict[VaspObject, VolumetricData]:
    """
    Load volumetric data files from a directory.

    Parameters
    ----------
    dir_name
        The directory containing the files.
    output_file_paths
        A dictionary mapping the data type to file path relative to dir_name.
    store_volumetric_data
        The volumetric data files to load. E.g., `("chgcar", "locpot")`.
        Provided as a list of strings note you can use either the keys or the
        values available in the `VaspObject` enum (e.g., "locpot" or "LOCPOT")
        are both valid.

    Returns
    -------
    dict[VaspObject, VolumetricData]
        A dictionary mapping the VASP object data type (`VaspObject.LOCPOT`,
        `VaspObject.CHGCAR`, etc) to the volumetric data object.
    """
    from pymatgen.io.vasp import Chgcar

    if store_volumetric_data is None or len(store_volumetric_data) == 0:
        return {}

    volumetric_data: dict[VaspObject, VolumetricData] = {}
    for file_type, file in output_file_paths.items():
        if (
            file_type.name not in store_volumetric_data
            and file_type.value not in store_volumetric_data
        ):
            continue

        try:
            # assume volumetric data is all in CHGCAR format
            volumetric_data[file_type] = Chgcar.from_file(str(dir_name / file))
        except Exception:
            raise ValueError(f"Failed to parse {file_type} at {file}.")
    return volumetric_data


def _parse_dos(parse_mode: str | bool, vasprun: Vasprun) -> Dos | None:
    """Parse DOS. See Calculation.from_vasp_files for supported arguments."""
    nsw = vasprun.incar.get("NSW", 0)
    dos = None
    if parse_mode is True or (parse_mode == "auto" and nsw < 1):
        dos = vasprun.complete_dos
    return dos


def _parse_bandstructure(
    parse_mode: str | bool, vasprun: Vasprun
) -> BandStructure | None:
    """Parse band structure. See Calculation.from_vasp_files for supported arguments."""
    vasprun_file = vasprun.filename

    if parse_mode == "auto":
        if vasprun.incar.get("ICHARG", 0) > 10:
            # NSCF calculation
            bs_vrun = BSVasprun(vasprun_file, parse_projected_eigen=True)
            try:
                # try parsing line mode
                bs = bs_vrun.get_band_structure(line_mode=True, efermi="smart")
            except Exception:
                # treat as a regular calculation
                bs = bs_vrun.get_band_structure(efermi="smart")
        else:
            # Not a NSCF calculation
            bs_vrun = BSVasprun(vasprun_file, parse_projected_eigen=False)
            bs = bs_vrun.get_band_structure(efermi="smart")

        # only save the bandstructure if not moving ions
        if vasprun.incar.get("NSW", 0) <= 1:
            return bs

    elif parse_mode:
        # legacy line/True behavior for bandstructure_mode
        bs_vrun = BSVasprun(vasprun_file, parse_projected_eigen=True)
        bs = bs_vrun.get_band_structure(line_mode=parse_mode == "line", efermi="smart")
        return bs

    return None


def _get_band_props(
    complete_dos: CompleteDos, structure: Structure
) -> dict[str, dict[str, dict[str, float]]]:
    """
    Calculate band properties from a CompleteDos object and Structure.

    Parameters
    ----------
    complete_dos
        A CompleteDos object.
    structure
        a pymatgen Structure object.

    Returns
    -------
    dict
        A dictionary of element and orbital-projected DOS properties.
    """
    dosprop_dict: dict[str, dict[str, dict[str, float]]] = {}
    if not complete_dos.pdos:
        # It's possible to have a CompleteDos object with only structure info and no projected DOS info
        return dosprop_dict

    for el in structure.composition.elements:
        el_name = str(el.name)
        dosprop_dict[el_name] = {}

        for orb_type in [
            OrbitalType(x) for x in range(OrbitalType[el.block].value + 1)  # type: ignore[misc]
        ]:
            try:
                dosprop_dict[el_name][str(orb_type)] = {
                    "filling": complete_dos.get_band_filling(
                        band=orb_type, elements=[el]
                    ),
                    "center": complete_dos.get_band_center(
                        band=orb_type, elements=[el]
                    ),
                    "bandwidth": complete_dos.get_band_width(
                        band=orb_type, elements=[el]
                    ),
                    "skewness": complete_dos.get_band_skewness(
                        band=orb_type, elements=[el]
                    ),
                    "kurtosis": complete_dos.get_band_kurtosis(
                        band=orb_type, elements=[el]
                    ),
                    "upper_edge": complete_dos.get_upper_band_edge(
                        band=orb_type, elements=[el]
                    ),
                }
            except KeyError:
                # No projected DOS available for that particular element + orbital character
                continue

    return dosprop_dict



import pytest

from monty.serialization import loadfn

from emmet.core.qchem.molecule import MoleculeDoc
from emmet.core.molecules.atomic import PartialChargesDoc, PartialSpinsDoc
from emmet.core.molecules.bonds import MoleculeBondingDoc
from emmet.core.molecules.thermo import MoleculeThermoDoc
from emmet.core.molecules.metal_binding import MetalBindingDoc


@pytest.fixture(scope="session")
def base_mol(test_dir):
    mol = loadfn((test_dir / "metal_binding_doc" / "base_mol_doc.json.gz").as_posix())
    mol_doc = MoleculeDoc(**mol)
    return mol_doc


@pytest.fixture(scope="session")
def base_thermo(test_dir):
    thermo = loadfn((test_dir / "metal_binding_doc" / "thermo.json.gz").as_posix())
    thermo_doc = MoleculeThermoDoc(**thermo)
    return thermo_doc


@pytest.fixture(scope="session")
def charges(test_dir):
    charges = loadfn(
        (test_dir / "metal_binding_doc" / "partial_charges.json.gz").as_posix()
    )
    charges_doc = PartialChargesDoc(**charges)
    return charges_doc


@pytest.fixture(scope="session")
def spins(test_dir):
    spins = loadfn(
        (test_dir / "metal_binding_doc" / "partial_spins.json.gz").as_posix()
    )
    spins_doc = PartialSpinsDoc(**spins)
    return spins_doc


@pytest.fixture(scope="session")
def bonds(test_dir):
    bonds = loadfn((test_dir / "metal_binding_doc" / "bonds.json.gz").as_posix())
    bonds_doc = MoleculeBondingDoc(**bonds)
    return bonds_doc


@pytest.fixture(scope="session")
def metal_thermo(test_dir):
    thermo = loadfn(
        (test_dir / "metal_binding_doc" / "metal_thermo.json.gz").as_posix()
    )
    metal_thermo_doc = MoleculeThermoDoc(**thermo)
    return metal_thermo_doc


@pytest.fixture(scope="session")
def nometal_thermo(test_dir):
    thermo = loadfn(
        (test_dir / "metal_binding_doc" / "nometal_thermo.json.gz").as_posix()
    )
    nometal_thermo_doc = MoleculeThermoDoc(**thermo)
    return nometal_thermo_doc


def test_metal_binding(
    base_mol, charges, spins, bonds, base_thermo, metal_thermo, nometal_thermo
):
    metal_binding = MetalBindingDoc.from_docs(
        method="mulliken-OB-mee",
        metal_indices=[5],
        base_molecule_doc=base_mol,
        partial_charges=charges,
        partial_spins=spins,
        bonding=bonds,
        base_thermo=base_thermo,
        metal_thermo={5: metal_thermo},
        nometal_thermo={5: nometal_thermo},
    )

    assert metal_binding.binding_data[0].metal_index == 5
    assert metal_binding.binding_data[0].metal_element == "Li"

    assert metal_binding.binding_data[0].metal_partial_charge == pytest.approx(
        -0.073376
    )
    assert metal_binding.binding_data[0].metal_assigned_charge == 0
    assert metal_binding.binding_data[0].metal_assigned_spin == 2
    assert metal_binding.binding_data[0].binding_free_energy == pytest.approx(
        1.23790290567376
    )



import pytest
from monty.serialization import loadfn

from emmet.core import ARROW_COMPATIBLE
from emmet.core.electronic_structure import ElectronicStructureDoc
from emmet.core.utils import jsanitize

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="session")
def structure(test_dir):
    """
    Fe (mp-13) structure with correct magmoms
    """
    structure = loadfn(test_dir / "electronic_structure/Fe_structure.json.gz")
    return structure


@pytest.fixture(scope="session")
def bandstructure(test_dir):
    """
    Fe (mp-13) line-mode band structure
    """
    bs = loadfn(test_dir / "electronic_structure/Fe_bs.json.gz")
    return bs


@pytest.fixture(scope="session")
def dos(test_dir):
    """
    Fe (mp-13) dos
    """
    dos = loadfn(test_dir / "electronic_structure/Fe_dos.json.gz")
    return dos


def test_from_bsdos_1(bandstructure, dos, structure):
    es_doc = ElectronicStructureDoc.from_bsdos(
        material_id="mp-13",
        dos={"mp-1671247": dos},
        is_gap_direct=False,
        is_metal=True,
        deprecated=False,
        setyawan_curtarolo={"mp-1056141": bandstructure},
        structures={"mp-1671247": structure, "mp-1056141": structure},
        meta_structure=structure,
    )

    assert es_doc.material_id == "mp-13"
    assert es_doc.band_gap == 0.0
    assert es_doc.cbm is None
    assert es_doc.vbm is None
    assert es_doc.efermi == 5.18804178
    assert es_doc.is_gap_direct is False
    assert es_doc.is_metal is True
    assert str(es_doc.magnetic_ordering) == "Ordering.FM"

    assert es_doc.bandstructure.setyawan_curtarolo.task_id == "mp-1056141"
    assert es_doc.bandstructure.setyawan_curtarolo.band_gap == 0.0
    assert es_doc.bandstructure.setyawan_curtarolo.efermi == 5.18804178
    assert es_doc.bandstructure.setyawan_curtarolo.nbands == 96.0


@pytest.fixture
def bandstructure_fs(test_dir):
    bs = loadfn(test_dir / "electronic_structure/es_bs_objs.json.gz")
    return bs


@pytest.fixture
def dos_fs(test_dir):
    dos = loadfn(test_dir / "electronic_structure/es_dos_objs.json.gz")
    return dos


def test_from_bsdos_2(bandstructure_fs, dos_fs):
    dos = dos_fs[0]["data"]
    bs = bandstructure_fs[0]["data"]

    es_doc = ElectronicStructureDoc.from_bsdos(
        material_id="mp-25375",
        dos={"mp-823888": dos},
        is_gap_direct=False,
        is_metal=True,
        deprecated=False,
        setyawan_curtarolo={"mp-1612487": bs},
        meta_structure=dos.structure,
    )

    assert es_doc.material_id == "mp-25375"
    assert es_doc.band_gap == pytest.approx(0.0)
    assert es_doc.cbm == pytest.approx(2.75448867)
    assert es_doc.vbm == pytest.approx(2.75448867)
    assert es_doc.efermi == pytest.approx(2.75448867)
    assert es_doc.is_gap_direct is False
    assert es_doc.is_metal is True
    assert str(es_doc.magnetic_ordering) == "Ordering.NM"

    assert es_doc.bandstructure.setyawan_curtarolo.task_id == "mp-1612487"
    assert es_doc.bandstructure.setyawan_curtarolo.band_gap == pytest.approx(1.9916)
    assert es_doc.bandstructure.setyawan_curtarolo.efermi == pytest.approx(2.49084067)
    assert es_doc.bandstructure.setyawan_curtarolo.nbands == pytest.approx(64.0)


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(bandstructure_fs, dos_fs):
    dos = dos_fs[0]["data"]
    bs = bandstructure_fs[0]["data"]

    doc = ElectronicStructureDoc.from_bsdos(
        material_id="mp-25375",
        dos={"mp-823888": dos},
        is_gap_direct=False,
        is_metal=True,
        deprecated=False,
        setyawan_curtarolo={"mp-1612487": bs},
        meta_structure=dos.structure,
    )

    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}),
        type=arrowize(ElectronicStructureDoc),
    )
    test_arrow_doc = ElectronicStructureDoc(
        **arrow_struct.as_py(maps_as_pydicts="strict")
    )

    # have to compare dicts due to dumping model to json the first time:
    #     DOS spins and orbital types don't rehydrate into pmg objects (Spin, OrbitalType)
    # str representation of the doesn't map back to the enum vals in pmg (ints)
    assert jsanitize(doc.model_dump(), allow_bson=True) == jsanitize(
        test_arrow_doc.model_dump(), allow_bson=True
    )
